# Copyright 2026 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from cwl_utils.parser.cwl_v1_2 import Workflow
from pathlib import Path
from types import SimpleNamespace
from typing import Union

from transpiler_mate import markdown as md


class _ArrayType:
    items = "string"


class _Template:
    def render(self, **kwargs):
        return (
            f"v={kwargs['version']};"
            f"wf={kwargs['workflow'].id};"
            f"name={kwargs['software_application']['name']}"
        )


def test_normalize_author_wraps_person_dict() -> None:
    roles = md.normalize_author(
        {
            "@type": "https://schema.org/Person",
            "given_name": "Ada",
            "family_name": "Lovelace",
        }
    )

    assert len(roles) == 1
    assert roles[0]["@type"] == "https://schema.org/Role"
    assert roles[0]["role_name"] == "N/A"
    assert roles[0]["author"]["@type"] == "https://schema.org/Person"


def test_normalize_contributor_flattens_nested_lists() -> None:
    roles = md.normalize_contributor(
        [
            {
                "@type": "https://schema.org/Person",
                "given_name": "Grace",
                "family_name": "Hopper",
            },
            [
                {
                    "contributor": {
                        "@type": "https://schema.org/Person",
                        "given_name": "Katherine",
                        "family_name": "Johnson",
                    }
                }
            ],
        ]
    )

    assert len(roles) == 2
    assert roles[0]["contributor"]["family_name"] == "Hopper"
    assert roles[1]["contributor"]["family_name"] == "Johnson"


def test_normalize_author_rejects_invalid_shapes() -> None:
    try:
        md.normalize_author({"unexpected": "shape"})
        assert False, "Expected ValueError"
    except ValueError as exc:
        assert "Unrecognized author dict shape" in str(exc)

    try:
        md.normalize_author(123)
        assert False, "Expected TypeError"
    except TypeError as exc:
        assert "author must be dict/list/None" in str(exc)


def test_type_to_string_and_nullable_helpers() -> None:
    parent = Workflow(inputs=[], outputs=[], steps=[])
    assert (
        md.type_to_string(Union[str, int], parent)
        == "One of:<ul><li>[str](https://www.commonwl.org/v1.2/Workflow.html#CWLType)</li><li>[int](https://www.commonwl.org/v1.2/Workflow.html#CWLType)</li></ul>"
    )
    assert (
        md.type_to_string(_ArrayType(), parent)
        == "`array` of [string](https://www.commonwl.org/v1.2/Workflow.html#CWLType)"
    )
    assert md.nullable(["null", "string"]) is True


def test_get_execution_command_formats_dynamic_arguments() -> None:
    command = md.get_exection_command(
        SimpleNamespace(
            baseCommand=["python", "-m", "module"],
            arguments=["--flag", 42],
        )
    )

    assert command == "python -m module --flag <ARGUMENT_DYNAMICALLY_SET>"


def test_markdown_transpile_renders_template(monkeypatch, tmp_path: Path) -> None:
    source = tmp_path / "workflow.cwl"
    source.write_text("cwlVersion: v1.2\n", encoding="utf-8")

    manager = SimpleNamespace(raw_document={"id": "#main"}, metadata=SimpleNamespace())

    monkeypatch.setattr(md, "MetadataManager", lambda _source: manager)
    monkeypatch.setattr(
        md,
        "CodeMetaTranspiler",
        lambda _repo: SimpleNamespace(
            transpile=lambda _metadata: {
                "@type": "SoftwareApplication",
                "name": "Example Tool",
            }
        ),
    )
    monkeypatch.setattr(md, "load_cwl_from_yaml", lambda _raw: "doc")
    monkeypatch.setattr(
        md, "search_process", lambda _id, _doc: SimpleNamespace(id="#main")
    )
    monkeypatch.setattr(md, "to_index", lambda _doc: {"main": "indexed"})
    monkeypatch.setattr(md, "_get_version", lambda: "9.9.9")
    monkeypatch.setattr(md.time, "time", lambda: 0)
    monkeypatch.setattr(
        md._jinja_environment, "get_template", lambda _name: _Template()
    )

    output = tmp_path / "out.md"
    with output.open("w", encoding="utf-8") as stream:
        md.markdown_transpile(source, "main", stream, None)

    rendered = output.read_text(encoding="utf-8")
    assert "v=9.9.9" in rendered
    assert "wf=#main" in rendered
    assert "name=Example Tool" in rendered


def test_markdown_transpile_raises_on_missing_workflow(
    monkeypatch, tmp_path: Path
) -> None:
    source = tmp_path / "workflow.cwl"
    source.write_text("cwlVersion: v1.2\n", encoding="utf-8")

    manager = SimpleNamespace(raw_document={"id": "#main"}, metadata=SimpleNamespace())

    monkeypatch.setattr(md, "MetadataManager", lambda _source: manager)
    monkeypatch.setattr(
        md,
        "CodeMetaTranspiler",
        lambda _repo: SimpleNamespace(
            transpile=lambda _metadata: {
                "@type": "SoftwareApplication",
                "name": "Example Tool",
            }
        ),
    )
    monkeypatch.setattr(md, "load_cwl_from_yaml", lambda _raw: "doc")
    monkeypatch.setattr(md, "search_process", lambda _id, _doc: None)

    try:
        with (tmp_path / "out.md").open("w", encoding="utf-8") as stream:
            md.markdown_transpile(source, "main", stream, None)
        assert False, "Expected ValueError"
    except ValueError as exc:
        assert "Workflow main does not exist" in str(exc)



from __future__ import annotations

from datetime import date
from httpx import Response
from pydantic import AnyUrl
from types import SimpleNamespace

from transpiler_mate import TranspilerBaseModel, _decode, _log_response
from transpiler_mate.invenio import (
    _affiliation_identifier,
    _to_creator,
    _to_identifier,
)
from transpiler_mate.metadata.software_application_models import (
    AuthorRole,
    CreativeWork,
    Organization,
    Person,
)
from transpiler_mate.oci import OrasAnnotationsTranspiler, _to_license_spdx
from invenio_rest_api_client.models.role_id import RoleId


class SampleModel(TranspilerBaseModel):
    homepage: AnyUrl


def _person() -> Person:
    return Person(
        givenName="Ada",
        familyName="Lovelace",
        email="ada@example.org",
        identifier="https://orcid.org/0000-0002-1825-0097",
        affiliation=Organization(
            name="Terradue",
            identifier="https://ror.org/03yrm5c26",
        ),
    )


def test_decode_supports_empty_string_and_bytes() -> None:
    assert _decode(None) == ""
    assert _decode("") == ""
    assert _decode("abc") == "abc"
    assert _decode(b"abc") == "abc"


def test_transpiler_base_model_dump_defaults_to_json_mode() -> None:
    model = SampleModel(homepage="https://example.org/path")
    dumped = model.model_dump()
    assert dumped["homepage"] == "https://example.org/path"


def test_log_response_raises_runtime_error_on_http_errors() -> None:
    def fake_request(*args, **kwargs):
        return Response(status_code=400, content=b"bad request")

    wrapped = _log_response(fake_request)

    try:
        wrapped(method="get", url="https://example.org")
        assert False, "Expected RuntimeError"
    except RuntimeError as exc:
        message = str(exc)
        assert "GET" in message
        assert "https://example.org" in message


def test_to_identifier_extracts_scheme_and_last_path_segment() -> None:
    identifier = _to_identifier("https://orcid.org/0000-0002-1825-0097")
    assert identifier.scheme.value == "orcid"
    assert identifier.identifier == "0000-0002-1825-0097"


def test_affiliation_identifier_returns_last_path_segment() -> None:
    assert _affiliation_identifier("https://ror.org/03yrm5c26") == "03yrm5c26"


def test_to_creator_from_person_builds_identifiers_and_affiliations() -> None:
    creator = _to_creator(_person())

    assert creator.role.id == RoleId.OTHER
    assert creator.person_or_org.name == "Lovelace, Ada"
    assert creator.person_or_org.identifiers[0].identifier == "0000-0002-1825-0097"
    assert creator.affiliations[0].id == "03yrm5c26"
    assert creator.affiliations[0].name == "Terradue"


def test_to_creator_from_author_role_maps_datacite_role() -> None:
    role = AuthorRole(
        roleName="Data Curator",
        additionalType="http://purl.org/spar/datacite/DataCurator",
        author=_person(),
    )

    creator = _to_creator(role)

    assert creator.role.id == RoleId.DATACURATOR


def test_to_license_spdx_supports_creative_work_and_url() -> None:
    assert _to_license_spdx(CreativeWork(identifier="Apache-2.0")) == "Apache-2.0"
    assert _to_license_spdx("https://spdx.org/licenses/MIT.html") == "MIT.html"


def test_oras_annotations_transpiler_generates_oci_annotations() -> None:
    metadata = SimpleNamespace(
        name="Example Tool",
        description="Example workflow metadata",
        software_version="1.2.3",
        date_created=date(2026, 3, 9),
        license=[
            CreativeWork(identifier="Apache-2.0"),
            "https://spdx.org/licenses/MIT.html",
        ],
    )
    process = SimpleNamespace(id="#main", cwlVersion="v1.2", class_="Workflow")

    annotations = OrasAnnotationsTranspiler(
        process=process,
        image_source="https://github.com/acme/example-tool",
        image_revision="abc123def",
    ).transpile(metadata)["$manifest"]

    assert annotations["org.opencontainers.image.title"] == "Example Tool"
    assert (
        annotations["org.opencontainers.image.description"]
        == "Example workflow metadata"
    )
    assert annotations["org.opencontainers.image.version"] == "1.2.3"
    assert (
        annotations["org.opencontainers.image.source"]
        == "https://github.com/acme/example-tool"
    )
    assert annotations["org.opencontainers.image.revision"] == "abc123def"
    assert annotations["org.opencontainers.image.licenses"] == "Apache-2.0 OR MIT.html"
    assert annotations["org.cwl.entrypoint"] == "#main"
    assert annotations["org.cwl.spec"] == "v1.2"
    assert annotations["org.cwl.type"] == "Workflow"



# Copyright 2026 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from datetime import date
from pathlib import Path

from transpiler_mate import metadata
from transpiler_mate.metadata.software_application_models import (
    CreativeWork,
    Organization,
    Person,
    SoftwareApplication,
)


class FakeYAML:
    def __init__(self):
        self.dumped_data = None

    def load(self, _source):
        return {"$namespaces": {}, "existing": "value"}

    def dump(self, data, stream):
        self.dumped_data = dict(data)
        stream.write("ok")


def _software_application() -> SoftwareApplication:
    publisher = Organization(name="Terradue")
    author = Person(
        givenName="Ada",
        familyName="Lovelace",
        email="ada@example.org",
        affiliation=publisher,
    )

    return SoftwareApplication(
        name="Example Tool",
        description="Example description",
        dateCreated=date(2026, 3, 9),
        license=CreativeWork(identifier="Apache-2.0"),
        softwareVersion="1.2.3",
        softwareHelp=CreativeWork(name="Help"),
        publisher=publisher,
        author=author,
    )


def test_metadata_manager_raises_for_missing_file(tmp_path: Path) -> None:
    missing = tmp_path / "missing.cwl"
    try:
        metadata.MetadataManager(missing)
        assert False, "Expected ValueError"
    except ValueError as exc:
        assert "non existing file" in str(exc)


def test_metadata_manager_raises_for_directory(tmp_path: Path) -> None:
    try:
        metadata.MetadataManager(tmp_path)
        assert False, "Expected ValueError"
    except ValueError as exc:
        assert "is not a file" in str(exc)


def test_metadata_manager_resolves_spdx_license_and_updates_document(
    monkeypatch, tmp_path: Path
) -> None:
    source = tmp_path / "workflow.cwl"
    source.write_text("dummy", encoding="utf-8")

    fake_yaml = FakeYAML()
    monkeypatch.setattr(metadata, "YAML", lambda: fake_yaml)

    def _compact(input_, ctx, options):
        if isinstance(input_, dict) and "existing" in input_:
            return input_
        return {"@context": "https://schema.org", "softwareVersion": "2.0.0"}

    monkeypatch.setattr(metadata.jsonld, "compact", _compact)

    app = _software_application()
    monkeypatch.setattr(
        metadata.SoftwareApplication,
        "model_validate",
        classmethod(lambda cls, compacted, by_alias=True: app),
    )

    manager = metadata.MetadataManager(source)

    # SPDX index should replace the lightweight CreativeWork with indexed metadata.
    assert isinstance(manager.metadata.license, CreativeWork)
    assert manager.metadata.license.identifier == "Apache-2.0"
    assert manager.metadata.license.name is not None

    manager.metadata.software_version = "2.0.0"
    manager.update()

    assert fake_yaml.dumped_data is not None
    assert fake_yaml.dumped_data["existing"] == "value"



# Copyright 2026 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from datetime import date

from transpiler_mate.codemeta import CodeMetaTranspiler
from transpiler_mate.metadata.software_application_models import (
    CreativeWork,
    DefinedTerm,
    Organization,
    Person,
    SoftwareApplication,
)


def _software_application() -> SoftwareApplication:
    publisher = Organization(name="Terradue")
    author = Person(
        givenName="Ada",
        familyName="Lovelace",
        email="ada@example.org",
        affiliation=publisher,
        identifier="https://orcid.org/0000-0002-1825-0097",
    )

    return SoftwareApplication(
        name="Example Tool",
        description="Example description",
        dateCreated=date(2026, 3, 9),
        license="https://spdx.org/licenses/Apache-2.0",
        softwareVersion="1.2.3",
        softwareHelp=CreativeWork(name="Help"),
        publisher=publisher,
        author=author,
        keywords=["remote-sensing", DefinedTerm(name="EO"), "cwl"],
    )


def test_codemeta_without_repository_keeps_application_and_filters_keywords() -> None:
    result = CodeMetaTranspiler(code_repository=None).transpile(_software_application())

    assert result["@context"] == "https://w3id.org/codemeta/3.0"
    assert result["@type"] == "SoftwareApplication"
    assert result["name"] == "Example Tool"
    assert result["keywords"] == ["remote-sensing", "cwl"]


def test_codemeta_with_github_repository_adds_repository_urls() -> None:
    result = CodeMetaTranspiler(
        code_repository="https://github.com/acme/example-tool.git"
    ).transpile(_software_application())

    assert result["@context"] == "https://w3id.org/codemeta/3.0"
    assert result["@type"] == "SoftwareSourceCode"
    assert result["codeRepository"] == "https://github.com/acme/example-tool.git"
    assert (
        result["continuousIntegration"]
        == "https://github.com/acme/example-tool/actions"
    )
    assert result["issueTracker"] == "https://github.com/acme/example-tool/issues"
    assert result["relatedLink"] == [
        "https://github.com/acme/example-tool/wiki",
        "https://github.com/acme/example-tool/releases",
        "https://github.com/acme/example-tool/deployments",
    ]



# Copyright 2026 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from datetime import date
from pathlib import Path
from types import SimpleNamespace

from invenio_rest_api_client.types import UNSET
from transpiler_mate.invenio import InvenioMetadataTranspiler, _md5
from transpiler_mate.metadata.software_application_models import (
    CreativeWork,
    Organization,
    Person,
    SoftwareApplication,
)


def _software_application(*, with_contributor: bool) -> SoftwareApplication:
    publisher = Organization(name="Terradue")
    author = Person(
        givenName="Ada",
        familyName="Lovelace",
        email="ada@example.org",
        affiliation=publisher,
        identifier="https://orcid.org/0000-0002-1825-0097",
    )

    contributor = Person(
        givenName="Grace",
        familyName="Hopper",
        email="grace@example.org",
        affiliation=publisher,
    )

    return SoftwareApplication(
        name="Example Tool",
        description="Example description",
        dateCreated=date(2026, 3, 9),
        license=CreativeWork(identifier="Apache-2.0"),
        softwareVersion="1.2.3",
        softwareHelp=CreativeWork(name="Help"),
        publisher=publisher,
        author=author,
        contributor=contributor if with_contributor else None,
    )


def test_md5_hashes_file_content(tmp_path: Path) -> None:
    data_file = tmp_path / "data.bin"
    data_file.write_bytes(b"abc")

    assert _md5(data_file) == "900150983cd24fb0d6963f7d28e17f72"


def test_invenio_transpile_maps_core_fields_with_contributors() -> None:
    transpiler = object.__new__(InvenioMetadataTranspiler)

    metadata = transpiler.transpile(_software_application(with_contributor=True))

    assert metadata.title == "Example Tool"
    assert metadata.publisher == "Terradue"
    assert metadata.version == "1.2.3"
    assert len(metadata.creators) == 1
    assert len(metadata.contributors) == 1


def test_invenio_transpile_sets_contributors_to_unset_when_missing() -> None:
    transpiler = object.__new__(InvenioMetadataTranspiler)

    metadata = transpiler.transpile(_software_application(with_contributor=False))

    assert metadata.contributors == UNSET


def test_create_or_update_process_reserves_doi_when_identifier_missing(
    monkeypatch, tmp_path: Path
) -> None:
    source = tmp_path / "workflow.cwl"
    attach = tmp_path / "readme.txt"
    source.write_text("class: Workflow\n", encoding="utf-8")
    attach.write_text("x", encoding="utf-8")

    metadata = SimpleNamespace(identifier=None, same_as=None)
    manager = SimpleNamespace(metadata=metadata, updated=False)
    manager.update = lambda: setattr(manager, "updated", True)

    class _SessionClient:
        def __enter__(self):
            return "session"

        def __exit__(self, exc_type, exc, tb):
            return False

    transpiler = object.__new__(InvenioMetadataTranspiler)
    transpiler.metadata_manager = manager
    transpiler.invenio_client = _SessionClient()
    transpiler.invenio_base_url = "https://invenio.example.org"
    transpiler.transpile = lambda _metadata: {"ok": True}

    captured = {}

    monkeypatch.setattr(
        "transpiler_mate.invenio.create_a_draft_record",
        lambda client, body: SimpleNamespace(to_dict=lambda: {"id": "42"}),
    )
    monkeypatch.setattr(
        "transpiler_mate.invenio.reserve_a_doi",
        lambda draft_id, client: SimpleNamespace(
            to_dict=lambda: {
                "doi": "10.1234/test.1",
                "doi_url": "https://doi.org/10.1234/test.1",
            }
        ),
    )
    monkeypatch.setattr(
        transpiler,
        "_finalize",
        lambda draft_id, uploading_files, session_client, invenio_metadata: (
            captured.update(
                {
                    "draft_id": draft_id,
                    "files": uploading_files,
                    "session_client": session_client,
                    "metadata": invenio_metadata,
                }
            )
            or "https://invenio.example.org/records/42"
        ),
    )

    url = transpiler.create_or_update_process(source=source, attach=(attach,))

    assert url == "https://invenio.example.org/records/42"
    assert metadata.identifier == "10.1234/test.1"
    assert str(metadata.same_as) == "https://doi.org/10.1234/test.1"
    assert manager.updated is True
    assert captured["draft_id"] == "42"
    assert captured["files"] == [source, attach]


def test_create_or_update_process_creates_new_version_for_existing_identifier(
    monkeypatch, tmp_path: Path
) -> None:
    source = tmp_path / "workflow.cwl"
    source.write_text("class: Workflow\n", encoding="utf-8")

    metadata = SimpleNamespace(identifier="10.1234/example.8", same_as=None)
    manager = SimpleNamespace(metadata=metadata, updated=False)
    manager.update = lambda: setattr(manager, "updated", True)

    class _SessionClient:
        def __enter__(self):
            return "session"

        def __exit__(self, exc_type, exc, tb):
            return False

    transpiler = object.__new__(InvenioMetadataTranspiler)
    transpiler.metadata_manager = manager
    transpiler.invenio_client = _SessionClient()
    transpiler.invenio_base_url = "https://invenio.example.org"
    transpiler.transpile = lambda _metadata: {"ok": True}

    monkeypatch.setattr(
        "transpiler_mate.invenio.create_a_new_version",
        lambda record_id, client: SimpleNamespace(to_dict=lambda: {"id": "77"}),
    )
    monkeypatch.setattr(
        transpiler,
        "_finalize",
        lambda draft_id, uploading_files, session_client, invenio_metadata: (
            f"https://invenio.example.org/records/{draft_id}"
        ),
    )

    url = transpiler.create_or_update_process(source=source, attach=None)

    assert url == "https://invenio.example.org/records/77"
    assert manager.updated is False


def test_finalize_uploads_files_and_publishes(monkeypatch, tmp_path: Path) -> None:
    source = tmp_path / "workflow.cwl"
    attach = tmp_path / "asset.txt"
    source.write_text("cwlVersion: v1.2\n", encoding="utf-8")
    attach.write_text("asset", encoding="utf-8")

    transpiler = object.__new__(InvenioMetadataTranspiler)
    transpiler.invenio_base_url = "https://invenio.example.org"

    calls = {
        "start": 0,
        "upload": 0,
        "complete": 0,
        "update": 0,
        "publish": 0,
    }

    monkeypatch.setattr(
        "transpiler_mate.invenio.step_1_start_draft_file_uploads",
        lambda draft_id, client, body: calls.__setitem__("start", calls["start"] + 1),
    )
    monkeypatch.setattr(
        "transpiler_mate.invenio.step_2_upload_a_draft_files_content",
        lambda draft_id, file_name, body, client: calls.__setitem__(
            "upload", calls["upload"] + 1
        ),
    )
    monkeypatch.setattr(
        "transpiler_mate.invenio.step_3_complete_a_draft_file_upload",
        lambda draft_id, file_name, client: calls.__setitem__(
            "complete", calls["complete"] + 1
        ),
    )
    monkeypatch.setattr(
        "transpiler_mate.invenio.update_a_draft_record",
        lambda draft_id, body, client: calls.__setitem__("update", calls["update"] + 1),
    )
    monkeypatch.setattr(
        "transpiler_mate.invenio.publish_a_draft_record",
        lambda draft_id, client: calls.__setitem__("publish", calls["publish"] + 1),
    )

    record_url = transpiler._finalize(
        draft_id="12",
        uploading_files=[source, attach],
        session_client="session",
        invenio_metadata={"title": "Example"},
    )

    assert record_url == "https://invenio.example.org/records/12"
    assert calls == {"start": 1, "upload": 2, "complete": 2, "update": 1, "publish": 1}



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .ogcapi_records_models import (
    Concept,
    Contact1,
    Email,
    Format1,
    Language,
    Link,
    RecordCommonProperties,
    RecordGeoJSON,
    Theme,
    Type7,
)
from .sciencekeywords import KEYWORDS_INDEX, ScienceKeywordRecord
from ..metadata.software_application_models import (
    AuthorRole,
    CreativeWork,
    DefinedTerm,
    Person,
    SoftwareApplication,
)
from ..metadata import Transpiler
from datetime import date, datetime, timezone
from loguru import logger

from pydantic import AnyUrl
from typing import Any, Mapping, List

import time
import uuid

SCIENCE_KEYWORDS_TERM_SET = AnyUrl(
    "https://gcmd.earthdata.nasa.gov/kms/concepts/concept_scheme/sciencekeywords"
)

DEFAULT_LANGUAGE: Language = Language(
    code="en-US", name="English (United States)", dir=None
)


def _to_datetime(value: date | datetime):
    if isinstance(value, datetime):
        return value if value.tzinfo else value.replace(tzinfo=timezone.utc)
    return datetime.combine(value, datetime.min.time(), tzinfo=timezone.utc)


def _to_contact(author: Person | AuthorRole) -> Contact1:
    position = None

    if isinstance(author, AuthorRole):
        position = author.role_name
        author = author.author

    return Contact1(
        identifier=str(author.identifier),
        name=f"{author.family_name}, {author.given_name}",
        organization=author.affiliation.name,
        position=position,
        emails=[Email(value=author.email)],
    )


class OgcRecordsTranspiler(Transpiler):
    def transpile(self, metadata_source: SoftwareApplication) -> Mapping[str, Any]:
        keywords: List[str] = []
        themes: List[Theme] = []

        if metadata_source.keywords:
            for raw_keyword in (
                metadata_source.keywords
                if isinstance(metadata_source.keywords, list)
                else [metadata_source.keywords]
            ):
                if isinstance(raw_keyword, str):
                    keywords.append(raw_keyword)
                elif isinstance(raw_keyword, DefinedTerm):
                    if (
                        SCIENCE_KEYWORDS_TERM_SET == raw_keyword.in_defined_term_set
                        and raw_keyword.term_code
                    ):
                        if raw_keyword.term_code not in KEYWORDS_INDEX:
                            logger.warning(
                                f"Science Keyword UUID {raw_keyword.term_code} not found in the index"
                            )
                        else:
                            science_keyword_record: ScienceKeywordRecord = (
                                KEYWORDS_INDEX[str(raw_keyword.term_code)]
                            )

                            concepts: List[Concept] = []

                            for i, keyword in enumerate(
                                science_keyword_record.hierarchy_list
                            ):
                                concepts.append(
                                    Concept(
                                        id=keyword,
                                        title=keyword,
                                        description=" > ".join(
                                            science_keyword_record.hierarchy_list[
                                                0 : i + 1
                                            ]
                                        ),
                                        url=science_keyword_record.uri,
                                    )
                                )

                            themes.append(
                                Theme(
                                    scheme=str(SCIENCE_KEYWORDS_TERM_SET),
                                    concepts=concepts,
                                )
                            )
                    else:
                        logger.debug(f"Discarding keyword, {raw_keyword}, unsupported")
                else:
                    logger.debug(
                        f"Discarding keyword, {raw_keyword}, unsupported type {type(raw_keyword)}"
                    )

        record_geojson: RecordGeoJSON = RecordGeoJSON(
            id=f"urn:uuid:{uuid.uuid4()}",
            type=Type7.FEATURE,
            links=list(
                map(
                    lambda creative_work: Link(
                        href=str(creative_work.url),
                        hreflang="en",
                        title=creative_work.name,
                        rel="help",
                        created=_to_datetime(metadata_source.date_created),
                        updated=_to_datetime(datetime.fromtimestamp(time.time())),
                    ),
                    metadata_source.software_help
                    if isinstance(metadata_source.software_help, list)
                    else [metadata_source.software_help],
                )
            ),
            properties=RecordCommonProperties(
                created=_to_datetime(metadata_source.date_created),
                updated=_to_datetime(datetime.fromtimestamp(time.time())),
                title=metadata_source.name,
                description=metadata_source.description
                if metadata_source.description
                else None,
                keywords=keywords if keywords else None,
                themes=themes if themes else None,
                language=DEFAULT_LANGUAGE,
                resource_languages=[DEFAULT_LANGUAGE],
                formats=[Format1(name="CWL", media_type="application/cwl")],
                contacts=list(
                    map(
                        _to_contact,
                        metadata_source.author
                        if isinstance(metadata_source.author, list)
                        else [metadata_source.author],
                    )
                ),
                license=": ".join(
                    list(
                        map(
                            lambda license: (
                                str(license.identifier)
                                if isinstance(license, CreativeWork)
                                else str(license)
                            ),
                            metadata_source.license
                            if isinstance(metadata_source.license, list)
                            else [metadata_source.license],
                        )
                    )
                ),
            ),
        )

        return record_geojson.model_dump(by_alias=True, exclude_none=True)



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from functools import wraps
from http import HTTPStatus
from httpx import Client, Headers, Request, RequestNotRead, Response
from pydantic import BaseModel, ConfigDict
from loguru import logger
from typing import Mapping
import re


def _decode(value):
    if not value:
        return ""

    if isinstance(value, str):
        return value

    return value.decode("utf-8")


def _log_request(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        request: Request = func(*args, **kwargs)

        logger.warning(f"{request.method} {request.url}")

        headers: Headers = request.headers
        for name, value in headers.raw:
            header_value = re.sub(
                r"(\bBearer\s+)[^\s]+",
                r"\1********",
                _decode(value),
                flags=re.IGNORECASE,
            )
            logger.warning(f"> {_decode(name)}: {header_value}")

        logger.warning(">")
        try:
            if request.content:
                logger.warning(_decode(request.content))
        except RequestNotRead:
            logger.warning("[REQUEST BUILT FROM STREAM, OMISSING]")

        return request

    return wrapper


def _log_response(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        response: Response = func(*args, **kwargs)

        if HTTPStatus.MULTIPLE_CHOICES._value_ <= response.status_code:
            log = logger.error
        else:
            log = logger.success

        status: HTTPStatus = HTTPStatus(response.status_code)
        log(f"< {status._value_} {status.phrase}")

        headers: Mapping[str, str] = response.headers
        for name, value in headers.items():
            log(f"< {_decode(name)}: {_decode(value)}")

        log("")

        if response.content:
            log(_decode(response.content))

        if HTTPStatus.MULTIPLE_CHOICES._value_ <= response.status_code:
            raise RuntimeError(
                f"A server error occurred when invoking {kwargs['method'].upper()} {kwargs['url']}, read the logs for details"
            )
        return response

    return wrapper


def init_http_logging(http_client: Client):
    http_client.build_request = _log_request(http_client.build_request)  # type: ignore
    http_client.request = _log_response(http_client.request)  # type: ignore


class TranspilerBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_by_name=True,
        extra="allow",
        use_enum_values=True,  # Enums dump as .value everywhere
    )

    # Default dumps to JSON-friendly types (URLs -> str, datetimes -> ISO, etc.)
    def model_dump(self, *args, **kwargs):
        kwargs.setdefault("mode", "json")
        kwargs.setdefault("exclude_none", True)
        return super().model_dump(*args, **kwargs)

    def model_dump_json(self, *args, **kwargs):
        # stays consistent with model_dump default
        kwargs.setdefault("indent", None)
        return super().model_dump_json(*args, **kwargs)



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from transpiler_mate.metadata.software_application_models import (
    SoftwareApplication,
    SoftwareSourceCode,
)
from .metadata import Transpiler

from giturlparse import parse as gitparse
from loguru import logger
from pydantic import BaseModel
from pyld import jsonld
from typing import Any, List, Mapping, MutableMapping


class CodeMetaTranspiler(Transpiler):
    def __init__(self, code_repository: str | None) -> None:
        self.code_repository = code_repository

    def transpile(self, metadata_source: SoftwareApplication) -> Mapping[str, Any]:
        model: BaseModel = metadata_source

        if self.code_repository:
            parsed_url = gitparse(self.code_repository)

            continuous_integration: str | None = None
            issue_tracker: str | None = None
            related_link: List[str] | None = None

            match parsed_url.platform:
                case "github":
                    continuous_integration = parsed_url.url2https.replace(
                        ".git", "/actions"
                    )
                    issue_tracker = parsed_url.url2https.replace(".git", "/issues")
                    related_link = [
                        parsed_url.url2https.replace(".git", page)
                        for page in ["/wiki", "/releases", "/deployments"]
                    ]

                case "gitlab":
                    continuous_integration = parsed_url.url2https.replace(
                        ".git", "/-/pipelines"
                    )
                    issue_tracker = parsed_url.url2https.replace(".git", "/-/issues")
                    related_link = [
                        parsed_url.url2https.replace(".git", page)
                        for page in ["/-/wikis/home", "/-/packages", "/-/pipelines"]
                    ]

                case _:
                    logger.warning(f"Platform {parsed_url.platform} unsupported yet")

            model = SoftwareSourceCode(
                code_repository=parsed_url.url2https,
                target_product=metadata_source,
                continuous_integration=continuous_integration,
                issue_tracker=issue_tracker,
                related_link=related_link,
            )  # type: ignore @type is a constant

        doc: MutableMapping[str, Any] = model.model_dump(
            exclude_none=True, by_alias=True
        )

        compacted: MutableMapping[str, Any] = jsonld.compact(
            doc,
            {
                "@context": {
                    "@vocab": "https://schema.org/",
                    # (optional) If you want relative IRIs to stay relative, omit @base.
                    # If you want to forbid a base so @id values don't get resolved, set:
                    # "@base": None,
                }
            },
            options={"processingMode": "json-ld-1.1", "ordered": None},
        )  # type: ignore

        compacted["@context"] = "https://w3id.org/codemeta/3.0"

        if metadata_source.keywords and isinstance(metadata_source.keywords, list):
            compacted["keywords"] = list(
                filter(
                    lambda keyword: isinstance(keyword, str), metadata_source.keywords
                )
            )

        return compacted



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .software_application_models import CreativeWork, SoftwareApplication
from .licenses import LICENSES_INDEX
from abc import abstractmethod
from loguru import logger
from pathlib import Path
from pydantic import AnyUrl
from pyld import jsonld
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap
from typing import Any, Generic, MutableMapping, TextIO, TypeVar

T = TypeVar("T")


class Transpiler(Generic[T]):
    @abstractmethod
    def transpile(self, metadata_source: SoftwareApplication) -> T:
        pass


__CONTEXT_KEY__ = "@context"
__NAMESPACES_KEY__ = "$namespaces"


class MetadataManager:
    def __init__(self, document_source: str | Path):
        if isinstance(document_source, str):
            document_source = Path(document_source)

        if not document_source.exists():
            raise ValueError(
                f"Input source document {document_source} points to a non existing file."
            )
        if not document_source.is_file():
            raise ValueError(f"Input source document {document_source} is not a file.")

        logger.debug(f"Loading raw document from {document_source}...")

        self.document_source = document_source
        self.yaml = YAML()

        self.raw_document: CommentedMap = self.yaml.load(document_source)

        compacted = jsonld.compact(
            input_=self.raw_document,
            ctx={},
            options={"expandContext": self.raw_document.get(__NAMESPACES_KEY__)},
        )
        self.metadata: SoftwareApplication = SoftwareApplication.model_validate(
            compacted, by_alias=True
        )

        logger.info("Resolving License details from SPDX License List...")

        def resolve_license(
            license: AnyUrl | CreativeWork | str,
        ) -> CreativeWork | AnyUrl:
            if isinstance(license, CreativeWork):
                return (
                    resolve_license(license.identifier)
                    if license.identifier and license.identifier in LICENSES_INDEX
                    else resolve_license(license.url)
                    if license.url and str(license.url) in LICENSES_INDEX
                    else license
                )
            elif isinstance(license, AnyUrl):
                return (
                    resolve_license(str(license))
                    if str(license) in LICENSES_INDEX
                    else license
                )

            resolved_license = LICENSES_INDEX[str(license)]
            return resolved_license

        if isinstance(self.metadata.license, list):
            for i, license in enumerate(self.metadata.license):
                if isinstance(license, CreativeWork):
                    self.metadata.license[i] = resolve_license(license)
        else:
            self.metadata.license = resolve_license(self.metadata.license)

    def update(self):
        metadata_dict = self.metadata.model_dump(exclude_none=True, by_alias=True)

        namespaces = self.raw_document.get(__NAMESPACES_KEY__)

        updated_metadata: MutableMapping[str, Any] = jsonld.compact(
            input_=metadata_dict, ctx=namespaces, options={"expandContext": namespaces}
        )  # type: ignore

        updated_metadata.pop(
            "@context"
        )  # remove undesired keys, $namespace already in the source document

        self.raw_document.update(updated_metadata)

        def _dump(stream: TextIO):
            self.yaml.dump(data=self.raw_document, stream=stream)

        logger.debug(
            "JSON-LD format compacted metadata merged to the original document"
        )
        with self.document_source.open("w") as output_stream:
            _dump(output_stream)

        logger.info(
            f"JSON-LD format compacted metadata merged to the original '{self.document_source}' document"
        )



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import init_http_logging
from .metadata import MetadataManager, Transpiler
from .metadata.software_application_models import (
    AuthorRole,
    ContributorRole,
    Person,
    Role as SWARole,
    SoftwareApplication,
)
from datetime import date

# required when a DOI is not assigned to an applicatrion package
from invenio_rest_api_client.client import AuthenticatedClient as InvenioClient
from invenio_rest_api_client.api.drafts.reserve_a_doi import sync as reserve_a_doi

from invenio_rest_api_client.api.records.create_a_draft_record import (
    sync as create_a_draft_record,
)
from invenio_rest_api_client.models.create_a_draft_record_body import (
    CreateADraftRecordBody,
)
from invenio_rest_api_client.models.created import Created

from invenio_rest_api_client.api.drafts_files_upload.step_1_start_draft_file_uploads import (
    sync as step_1_start_draft_file_uploads,
)
from invenio_rest_api_client.models.file_transfer_item import FileTransferItem

from invenio_rest_api_client.api.drafts_files_upload.step_2_upload_a_draft_files_content import (
    sync as step_2_upload_a_draft_files_content,
)
from invenio_rest_api_client.types import File as FileContent

from invenio_rest_api_client.api.drafts_files_upload.step_3_complete_a_draft_file_upload import (
    sync as step_3_complete_a_draft_file_upload,
)

from invenio_rest_api_client.api.drafts.update_a_draft_record import (
    sync as update_a_draft_record,
)
from invenio_rest_api_client.models.access import Access
from invenio_rest_api_client.models.access_files import AccessFiles
from invenio_rest_api_client.models.access_record import AccessRecord
from invenio_rest_api_client.models.affiliation import Affiliation
from invenio_rest_api_client.models.creator import Creator
from invenio_rest_api_client.models.files import Files
from invenio_rest_api_client.models.identifier import Identifier
from invenio_rest_api_client.models.person_or_org_identifier_scheme import (
    PersonOrOrgIdentifierScheme,
)
from invenio_rest_api_client.models.metadata import Metadata
from invenio_rest_api_client.models.person_or_org import PersonOrOrg
from invenio_rest_api_client.models.person_or_org_type import PersonOrOrgType
from invenio_rest_api_client.models.resource_type import ResourceType
from invenio_rest_api_client.models.resource_type_id import ResourceTypeId
from invenio_rest_api_client.models.role import Role
from invenio_rest_api_client.models.role_id import RoleId
from invenio_rest_api_client.models.update_draft_record import UpdateDraftRecord

from invenio_rest_api_client.api.drafts.publish_a_draft_record import (
    sync as publish_a_draft_record,
)

from invenio_rest_api_client.api.records_versions.create_a_new_version import (
    sync as create_a_new_version,
)

from invenio_rest_api_client.types import UNSET

from loguru import logger
from pathlib import Path
from pydantic import AnyUrl
from typing import List, Mapping, Optional, Tuple
from urllib.parse import urlparse

import hashlib
import time

__ROLES_MAPPING_: Mapping[AnyUrl, RoleId] = {
    AnyUrl("http://purl.org/spar/datacite/ContactPerson"): RoleId.CONTACTPERSON,
    AnyUrl("http://purl.org/spar/datacite/DataCollector"): RoleId.DATACOLLECTOR,
    AnyUrl("http://purl.org/spar/datacite/DataCurator"): RoleId.DATACURATOR,
    AnyUrl("http://purl.org/spar/datacite/DataManager"): RoleId.DATAMANAGER,
    AnyUrl("http://purl.org/spar/datacite/Distributor"): RoleId.DISTRIBUTOR,
    AnyUrl("http://purl.org/spar/datacite/Editor"): RoleId.EDITOR,
    AnyUrl(
        "http://purl.org/spar/datacite/HostingInstitution"
    ): RoleId.HOSTINGINSTITUTION,
    AnyUrl("http://purl.org/spar/datacite/Other"): RoleId.OTHER,
    AnyUrl("http://purl.org/spar/datacite/Producer"): RoleId.PRODUCER,
    AnyUrl("http://purl.org/spar/datacite/ProjectLeader"): RoleId.PROJECTLEADER,
    AnyUrl("http://purl.org/spar/datacite/ProjectManager"): RoleId.PROJECTMANAGER,
    AnyUrl("http://purl.org/spar/datacite/ProjectMember"): RoleId.PROJECTMEMBER,
    AnyUrl(
        "http://purl.org/spar/datacite/RegistrationAgency"
    ): RoleId.REGISTRATIONAGENCY,
    AnyUrl(
        "http://purl.org/spar/datacite/RegistrationAuthority"
    ): RoleId.REGISTRATIONAUTHORITY,
    AnyUrl("http://purl.org/spar/datacite/RelatedPerson"): RoleId.RELATEDPERSON,
    AnyUrl("http://purl.org/spar/datacite/Researcher"): RoleId.RESEARCHER,
    AnyUrl("http://purl.org/spar/datacite/ResearchGroup"): RoleId.RESEARCHGROUP,
    AnyUrl("http://purl.org/spar/datacite/RightsHolder"): RoleId.RIGHTSHOLDER,
    AnyUrl("http://purl.org/spar/datacite/Sponsor"): RoleId.SPONSOR,
    AnyUrl("http://purl.org/spar/datacite/Supervisor"): RoleId.SUPERVISOR,
    AnyUrl("http://purl.org/spar/datacite/WorkPackageLeader"): RoleId.WORKPACKAGELEADER,
}


def _md5(file: Path):
    hash_md5 = hashlib.md5()
    with file.open("rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def _to_identifier(url_identifier: str | AnyUrl) -> Identifier:
    _, netloc, path, _, _, _ = urlparse(str(url_identifier))
    return Identifier(
        scheme=PersonOrOrgIdentifierScheme(netloc.split(".")[0]),
        identifier=path.split("/")[-1],
    )


def _affiliation_identifier(url_identifier: str | AnyUrl) -> str:
    _, _, path, _, _, _ = urlparse(str(url_identifier))
    return path.split("/")[-1]


def _to_creator(author: Person | SWARole) -> Creator:
    role_id: RoleId = RoleId.OTHER

    if isinstance(author, SWARole):
        if author.additional_type:
            role_id = __ROLES_MAPPING_.get(author.additional_type, RoleId.OTHER)

        if isinstance(author, AuthorRole):
            author = author.author
        elif isinstance(author, ContributorRole):
            author = author.contributor

    creator: Creator = Creator(
        person_or_org=PersonOrOrg(
            type_=PersonOrOrgType.PERSONAL,
            name=f"{author.family_name}, {author.given_name}"
            if isinstance(author, Person)
            else UNSET,
            given_name=author.given_name if isinstance(author, Person) else UNSET,
            family_name=author.family_name if isinstance(author, Person) else UNSET,
            identifiers=[_to_identifier(author.identifier)]
            if isinstance(author, Person) and author.identifier
            else UNSET,
        ),
        role=Role(id=role_id),
    )

    if isinstance(author, Person):
        creator.affiliations = []
        for affiliation in (
            author.affiliation
            if isinstance(author.affiliation, list)
            else [author.affiliation]
        ):
            creator.affiliations.append(
                Affiliation(
                    id=_affiliation_identifier(affiliation.identifier)
                    if affiliation.identifier
                    else UNSET,
                    name=affiliation.name,
                )
            )

    return creator


class InvenioMetadataTranspiler(Transpiler):
    def __init__(
        self, metadata_manager: MetadataManager, invenio_base_url: str, auth_token: str
    ):
        self.metadata_manager = metadata_manager

        self.invenio_base_url = invenio_base_url
        self.invenio_client: InvenioClient = InvenioClient(
            base_url=invenio_base_url, token=auth_token
        )

        logger.debug("Setting up the HTTP logger...")
        init_http_logging(self.invenio_client.get_httpx_client())
        logger.debug("HTTP logger correctly setup")

    def transpile(self, metadata_source: SoftwareApplication) -> Metadata:
        return Metadata(
            resource_type=ResourceType(id=ResourceTypeId.WORKFLOW),
            title=metadata_source.name,
            publication_date=date.fromtimestamp(time.time()),
            publisher=metadata_source.publisher.name,
            description=metadata_source.description
            if metadata_source.description
            else UNSET,
            creators=list(
                map(
                    _to_creator,
                    metadata_source.author
                    if isinstance(metadata_source.author, list)
                    else [metadata_source.author],
                )
            ),
            contributors=list(
                map(
                    _to_creator,
                    metadata_source.contributor
                    if isinstance(metadata_source.contributor, list)
                    else [metadata_source.contributor],
                )
            )
            if metadata_source.contributor
            else UNSET,
            version=metadata_source.software_version,
        )

    def _finalize(
        self,
        draft_id: str,
        uploading_files: List[Path],
        session_client: InvenioClient,
        invenio_metadata: Metadata,
    ) -> str:
        uploading_files_names = ", ".join([file.name for file in uploading_files])
        logger.info(
            f"Drafting file upload [{uploading_files_names}] to Record '{draft_id}'..."
        )

        step_1_start_draft_file_uploads(
            draft_id=draft_id,
            client=session_client,
            body=[
                FileTransferItem(
                    key=file.name,
                    size=file.stat().st_size,
                    checksum=f"md5:{_md5(file)}",
                )
                for file in uploading_files
            ],
        )

        logger.success(
            f"File upload {uploading_files_names} drafted to Record '{draft_id}'"
        )

        for file in uploading_files:
            logger.info(
                f"Uploading file content '{file.name})' to Record '{draft_id}'..."
            )

            with file.open("rb") as binary_stream:
                step_2_upload_a_draft_files_content(
                    draft_id=draft_id,
                    file_name=file.name,
                    body=FileContent(
                        file_name=file.name,
                        mime_type="application/octet-stream",
                        payload=binary_stream,
                    ),
                    client=session_client,
                )

            logger.success(f"File content {file.name} uploaded to Record {draft_id}")

            logger.info(
                f"Completing file upload {file.name}] to Record '{draft_id}'..."
            )

            step_3_complete_a_draft_file_upload(
                draft_id=draft_id, file_name=file.name, client=session_client
            )

            logger.success(f"File upload {file.name} to Record '{draft_id}' completed")

        update_a_draft_record(
            draft_id=draft_id,
            body=UpdateDraftRecord(
                access=Access(files=AccessFiles.PUBLIC, record=AccessRecord.PUBLIC),
                files=Files(enabled=True),
                metadata=invenio_metadata,
            ),
            client=session_client,
        )

        logger.success(f"Draft Record '{draft_id}' metadata updated!")

        logger.info(f"Publishing the Draft Record '{draft_id}'...")

        publish_a_draft_record(draft_id=draft_id, client=session_client)

        logger.success(f"Draft Record '{draft_id}' metadata updated!")

        return f"{self.invenio_base_url}/records/{draft_id}"

    def create_or_update_process(
        self, source: Path, attach: Optional[Tuple[Path]] = None
    ) -> str:
        metadata: SoftwareApplication = self.metadata_manager.metadata

        with self.invenio_client as invenio_rest_client:
            draft_id: str = ""

            if not metadata.identifier:
                logger.warning(
                    "'identifier' key not found in source document, reserving a DOI..."
                )

                draft_record = create_a_draft_record(
                    client=invenio_rest_client, body=CreateADraftRecordBody()
                )

                draft_id = (
                    draft_record.id
                    if draft_record and isinstance(draft_record, Created)
                    else draft_record.to_dict()["id"]
                    if draft_record
                    else ""
                )  # type: ignore

                logger.success(
                    f"Successfully reserved a draft record with ID: {draft_id}"
                )

                doi = reserve_a_doi(draft_id=draft_id, client=invenio_rest_client)

                doi_dict = doi.to_dict()  # type: ignore
                doi = doi_dict["doi"]
                doi_url = doi_dict["doi_url"]
                logger.success(
                    f"Successfully reserved a DOI with ID {doi} (URL: {doi_url})"
                )

                metadata.identifier = doi
                metadata.same_as = AnyUrl(doi_url)

                self.metadata_manager.update()
            else:
                logger.info(
                    f"Identifier {metadata.identifier} already assigned to {source}"
                )

                record_id = metadata.identifier.split(".")[-1]  # type: ignore

                logger.info(
                    f"Creating a new version for already existing Record {record_id}"
                )

                version = create_a_new_version(
                    record_id=record_id, client=invenio_rest_client
                )

                draft_id = version.to_dict()["id"]  # type: ignore

                logger.info(
                    f"New version {draft_id} for already existing Record {record_id} created!"
                )

            uploading_files = [source]
            if attach:
                for attach_item in attach:
                    uploading_files.append(attach_item)

            return self._finalize(
                draft_id=draft_id,
                uploading_files=uploading_files,
                session_client=invenio_rest_client,
                invenio_metadata=self.transpile(metadata),
            )



# Copyright 2026 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .oci_annotations_models import OciAnnotations
from ..metadata import Transpiler
from ..metadata.software_application_models import CreativeWork, SoftwareApplication
from cwl_utils.parser import Process
from pydantic import AnyUrl
from typing import Any, Mapping


def _to_license_spdx(license: CreativeWork | AnyUrl) -> str:
    if isinstance(license, CreativeWork):
        return str(license.identifier)
    return str(license).split("/")[-1]


class OrasAnnotationsTranspiler(Transpiler):
    def __init__(
        self,
        process: Process,
        image_source: str | None,
        image_revision: str | None,
    ):
        self.process: Process = process
        self.image_source = image_source
        self.image_revision = image_revision

    def transpile(self, metadata_source: SoftwareApplication) -> Mapping[str, Any]:
        oci_annotations: OciAnnotations = OciAnnotations()

        # org.opencontainers.image.* properties
        oci_annotations.org_opencontainers_image_title = metadata_source.name
        oci_annotations.org_opencontainers_image_description = (
            metadata_source.description
        )
        oci_annotations.org_opencontainers_image_version = (
            metadata_source.software_version
        )
        oci_annotations.org_opencontainers_image_source = self.image_source
        oci_annotations.org_opencontainers_image_revision = self.image_revision
        oci_annotations.org_opencontainers_image_licenses = (
            " OR ".join(
                [_to_license_spdx(license) for license in metadata_source.license]
            )
            if isinstance(metadata_source.license, list)
            else _to_license_spdx(metadata_source.license)
        )

        # org.cwl.* properties
        oci_annotations.org_cwl_entrypoint = self.process.id
        oci_annotations.org_cwl_spec = self.process.cwlVersion
        oci_annotations.org_cwl_type = self.process.class_

        return {
            "$manifest": oci_annotations.model_dump(by_alias=True, exclude_none=True)
        }



# Copyright 2025-2026 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# generated by datamodel-codegen:
#   filename:  software_application.yaml
#   timestamp: 2026-03-05T14:51:11+00:00

from __future__ import annotations

from datetime import date, datetime
from typing import List, Literal, Optional, Union

from pydantic import AliasChoices, AnyUrl, EmailStr, Field, RootModel
from transpiler_mate import TranspilerBaseModel


class Organization(TranspilerBaseModel):
    """
    An organization such as a school, NGO, corporation, club, etc.
    """

    field_type: Literal["https://schema.org/Organization"] = Field(
        "https://schema.org/Organization", alias="@type"
    )
    name: str = Field(
        ...,
        validation_alias=AliasChoices("name", "https://schema.org/name"),
        serialization_alias="https://schema.org/name",
    )
    email: Optional[Union[EmailStr, List[EmailStr]]] = Field(
        default=None,
        validation_alias=AliasChoices("email", "https://schema.org/email"),
        serialization_alias="https://schema.org/email",
    )
    identifier: Optional[Union[AnyUrl, str]] = Field(
        default=None,
        validation_alias=AliasChoices("identifier", "https://schema.org/identifier"),
        serialization_alias="https://schema.org/identifier",
    )


class Person(TranspilerBaseModel):
    """
    A person (alive, dead, undead, or fictional).
    """

    field_type: Literal["https://schema.org/Person"] = Field(
        "https://schema.org/Person", alias="@type"
    )
    given_name: str = Field(
        ...,
        validation_alias=AliasChoices("givenName", "https://schema.org/givenName"),
        serialization_alias="https://schema.org/givenName",
    )
    family_name: str = Field(
        ...,
        validation_alias=AliasChoices("familyName", "https://schema.org/familyName"),
        serialization_alias="https://schema.org/familyName",
    )
    email: Union[EmailStr, List[EmailStr]] = Field(
        ...,
        validation_alias=AliasChoices("email", "https://schema.org/email"),
        serialization_alias="https://schema.org/email",
    )
    identifier: Optional[Union[AnyUrl, str]] = Field(
        default=None,
        validation_alias=AliasChoices("identifier", "https://schema.org/identifier"),
        serialization_alias="https://schema.org/identifier",
    )
    affiliation: Union[Organization, List[Organization]] = Field(
        ...,
        validation_alias=AliasChoices("affiliation", "https://schema.org/affiliation"),
        serialization_alias="https://schema.org/affiliation",
    )


class Role(TranspilerBaseModel):
    """
    Represents additional information about a relationship or property.
    """

    field_type: Literal["https://schema.org/Role"] = Field(
        "https://schema.org/Role", alias="@type"
    )
    role_name: str = Field(
        ...,
        validation_alias=AliasChoices("roleName", "https://schema.org/roleName"),
        serialization_alias="https://schema.org/roleName",
    )
    start_date: Optional[date] = Field(
        default=None,
        validation_alias=AliasChoices("startDate", "https://schema.org/startDate"),
        serialization_alias="https://schema.org/startDate",
    )
    end_date: Optional[date] = Field(
        default=None,
        validation_alias=AliasChoices("endDate", "https://schema.org/endDate"),
        serialization_alias="https://schema.org/endDate",
    )
    additional_type: Optional[AnyUrl] = Field(
        default=None,
        validation_alias=AliasChoices(
            "additionalType", "https://schema.org/additionalType"
        ),
        serialization_alias="https://schema.org/additionalType",
    )


class AuthorRole(Role):
    author: Person = Field(
        ...,
        validation_alias=AliasChoices("author", "https://schema.org/author"),
        serialization_alias="https://schema.org/author",
    )


class ContributorRole(Role):
    contributor: Person = Field(
        ...,
        validation_alias=AliasChoices("contributor", "https://schema.org/contributor"),
        serialization_alias="https://schema.org/contributor",
    )


class DefinedTerm(TranspilerBaseModel):
    """
    A word, name, acronym, phrase, etc. with a formal definition. Often used in the context of category or subject classification, glossaries or dictionaries, product or creative work types, etc. Use the name property for the term being defined, use termCode if the term has an alpha-numeric code allocated, use description to provide the definition of the term.
    """

    field_type: Literal["https://schema.org/DefinedTerm"] = Field(
        "https://schema.org/DefinedTerm", alias="@type"
    )
    name: Optional[str] = Field(
        default=None,
        validation_alias=AliasChoices("name", "https://schema.org/name"),
        serialization_alias="https://schema.org/name",
    )
    description: Optional[str] = Field(
        default=None,
        validation_alias=AliasChoices("description", "https://schema.org/description"),
        serialization_alias="https://schema.org/description",
    )
    term_code: Optional[str] = Field(
        default=None,
        validation_alias=AliasChoices("termCode", "https://schema.org/termCode"),
        serialization_alias="https://schema.org/termCode",
    )
    in_defined_term_set: Optional[AnyUrl] = Field(
        default=None,
        validation_alias=AliasChoices(
            "inDefinedTermSet", "https://schema.org/inDefinedTermSet"
        ),
        serialization_alias="https://schema.org/inDefinedTermSet",
    )


class CreativeWork(TranspilerBaseModel):
    """
    The most generic kind of creative work, including books, movies, photographs, software programs, etc.
    """

    field_type: Literal["https://schema.org/CreativeWork"] = Field(
        "https://schema.org/CreativeWork", alias="@type"
    )
    name: Optional[str] = Field(
        default=None,
        validation_alias=AliasChoices("name", "https://schema.org/name"),
        serialization_alias="https://schema.org/name",
    )
    url: Optional[AnyUrl] = Field(
        default=None,
        validation_alias=AliasChoices("url", "https://schema.org/url"),
        serialization_alias="https://schema.org/url",
    )
    identifier: Optional[Union[AnyUrl, str]] = Field(
        default=None,
        validation_alias=AliasChoices("identifier", "https://schema.org/identifier"),
        serialization_alias="https://schema.org/identifier",
    )


class ImageObject(TranspilerBaseModel):
    """
    An image file.
    """

    field_type: Literal["https://schema.org/ImageObject"] = Field(
        "https://schema.org/ImageObject", alias="@type"
    )
    content_url: Optional[AnyUrl] = Field(
        default=None,
        validation_alias=AliasChoices("contentUrl", "https://schema.org/contentUrl"),
        serialization_alias="https://schema.org/contentUrl",
    )
    caption: Optional[str] = Field(
        default=None,
        validation_alias=AliasChoices("caption", "https://schema.org/caption"),
        serialization_alias="https://schema.org/caption",
    )
    encoding_format: Optional[str] = Field(
        default=None,
        validation_alias=AliasChoices(
            "encodingFormat", "https://schema.org/encodingFormat"
        ),
        serialization_alias="https://schema.org/encodingFormat",
    )
    width: Optional[int] = Field(
        default=None,
        validation_alias=AliasChoices("width", "https://schema.org/width"),
        serialization_alias="https://schema.org/width",
    )
    height: Optional[int] = Field(
        default=None,
        validation_alias=AliasChoices("height", "https://schema.org/height"),
        serialization_alias="https://schema.org/height",
    )


class SoftwareApplication(TranspilerBaseModel):
    """
    A software application.
    """

    field_type: Literal["https://schema.org/SoftwareApplication"] = Field(
        "https://schema.org/SoftwareApplication", alias="@type"
    )
    name: str = Field(
        ...,
        validation_alias=AliasChoices("name", "https://schema.org/name"),
        serialization_alias="https://schema.org/name",
    )
    description: str = Field(
        ...,
        validation_alias=AliasChoices("description", "https://schema.org/description"),
        serialization_alias="https://schema.org/description",
    )
    date_created: Union[date, datetime] = Field(
        ...,
        validation_alias=AliasChoices("dateCreated", "https://schema.org/dateCreated"),
        serialization_alias="https://schema.org/dateCreated",
    )
    license: Union[CreativeWork, AnyUrl, List[Union[CreativeWork, AnyUrl]]] = Field(
        ...,
        validation_alias=AliasChoices("license", "https://schema.org/license"),
        serialization_alias="https://schema.org/license",
    )
    identifier: Optional[Union[AnyUrl, str]] = Field(
        default=None,
        validation_alias=AliasChoices("identifier", "https://schema.org/identifier"),
        serialization_alias="https://schema.org/identifier",
    )
    same_as: Optional[Union[AnyUrl, List[AnyUrl]]] = Field(
        default=None,
        validation_alias=AliasChoices("sameAs", "https://schema.org/sameAs"),
        serialization_alias="https://schema.org/sameAs",
    )
    keywords: Optional[
        Union[str, AnyUrl, DefinedTerm, List[Union[str, AnyUrl, DefinedTerm]]]
    ] = Field(
        default=None,
        validation_alias=AliasChoices("keywords", "https://schema.org/keywords"),
        serialization_alias="https://schema.org/keywords",
    )
    operating_system: Optional[Union[str, List[str]]] = Field(
        default=None,
        validation_alias=AliasChoices(
            "operatingSystem", "https://schema.org/operatingSystem"
        ),
        serialization_alias="https://schema.org/operatingSystem",
    )
    software_requirements: Optional[Union[str, AnyUrl, List[Union[str, AnyUrl]]]] = (
        Field(
            default=None,
            validation_alias=AliasChoices(
                "softwareRequirements", "https://schema.org/softwareRequirements"
            ),
            serialization_alias="https://schema.org/softwareRequirements",
        )
    )
    software_version: str = Field(
        ...,
        validation_alias=AliasChoices(
            "softwareVersion", "https://schema.org/softwareVersion"
        ),
        serialization_alias="https://schema.org/softwareVersion",
    )
    software_help: Union[CreativeWork, List[CreativeWork]] = Field(
        ...,
        validation_alias=AliasChoices(
            "softwareHelp", "https://schema.org/softwareHelp"
        ),
        serialization_alias="https://schema.org/softwareHelp",
    )
    publisher: Organization = Field(
        ...,
        validation_alias=AliasChoices("publisher", "https://schema.org/publisher"),
        serialization_alias="https://schema.org/publisher",
    )
    author: Union[AuthorRole, Person, List[Union[AuthorRole, Person]]] = Field(
        ...,
        validation_alias=AliasChoices("author", "https://schema.org/author"),
        serialization_alias="https://schema.org/author",
    )
    contributor: Optional[
        Union[ContributorRole, Person, List[Union[ContributorRole, Person]]]
    ] = Field(
        default=None,
        validation_alias=AliasChoices("contributor", "https://schema.org/contributor"),
        serialization_alias="https://schema.org/contributor",
    )
    thumbnail: Optional[ImageObject] = Field(
        default=None,
        validation_alias=AliasChoices("thumbnail", "https://schema.org/thumbnail"),
        serialization_alias="https://schema.org/thumbnail",
    )


class SoftwareSourceCode(TranspilerBaseModel):
    """
    Computer programming source code.
    """

    field_type: Literal["https://schema.org/SoftwareSourceCode"] = Field(
        "https://schema.org/SoftwareSourceCode", alias="@type"
    )
    code_repository: AnyUrl = Field(
        ...,
        validation_alias=AliasChoices(
            "codeRepository", "https://schema.org/codeRepository"
        ),
        serialization_alias="https://schema.org/codeRepository",
    )
    target_product: SoftwareApplication = Field(
        ...,
        validation_alias=AliasChoices(
            "targetProduct", "https://schema.org/targetProduct"
        ),
        serialization_alias="https://schema.org/targetProduct",
    )
    continuous_integration: Optional[AnyUrl] = Field(
        default=None,
        validation_alias=AliasChoices(
            "continuousIntegration", "https://schema.org/continuousIntegration"
        ),
        serialization_alias="https://schema.org/continuousIntegration",
    )
    issue_tracker: Optional[AnyUrl] = Field(
        default=None,
        validation_alias=AliasChoices(
            "issueTracker", "https://schema.org/issueTracker"
        ),
        serialization_alias="https://schema.org/issueTracker",
    )
    related_link: Optional[Union[AnyUrl, List[AnyUrl]]] = Field(
        default=None,
        validation_alias=AliasChoices("relatedLink", "https://schema.org/relatedLink"),
        serialization_alias="https://schema.org/relatedLink",
    )


class Model(RootModel[SoftwareApplication]):
    root: SoftwareApplication



# Copyright 2026 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from datetime import date

from transpiler_mate.datacite import DataCiteTranspiler
from transpiler_mate.metadata.software_application_models import (
    AuthorRole,
    ContributorRole,
    CreativeWork,
    Organization,
    Person,
    SoftwareApplication,
)


def _software_application(*, with_identifier: bool) -> SoftwareApplication:
    publisher = Organization(name="Terradue")
    author = AuthorRole(
        roleName="Lead",
        author=Person(
            givenName="Ada",
            familyName="Lovelace",
            email="ada@example.org",
            identifier="https://orcid.org/0000-0002-1825-0097",
            affiliation=Organization(
                name="Terradue",
                identifier="https://ror.org/03yrm5c26",
            ),
        ),
    )
    contributor = ContributorRole(
        roleName="Data Curator",
        additionalType="http://purl.org/spar/datacite/DataCurator",
        contributor=Person(
            givenName="Grace",
            familyName="Hopper",
            email="grace@example.org",
            identifier="https://orcid.org/0000-0002-1825-0100",
            affiliation=Organization(
                name="Navy", identifier="https://ror.org/01aaaab11"
            ),
        ),
    )

    kwargs = {}
    if with_identifier:
        kwargs["identifier"] = "10.1234/example.1"
        kwargs["sameAs"] = "https://doi.org/10.1234/example.1"

    return SoftwareApplication(
        name="Example Tool",
        description="Example description",
        dateCreated=date(2026, 3, 9),
        license=CreativeWork(
            name="Apache License 2.0",
            url="https://spdx.org/licenses/Apache-2.0.html",
            identifier="Apache-2.0",
        ),
        softwareVersion="1.2.3",
        softwareHelp=CreativeWork(name="Help"),
        publisher=publisher,
        author=author,
        contributor=contributor,
        **kwargs,
    )


def test_datacite_transpile_maps_identifier_and_people() -> None:
    result = DataCiteTranspiler().transpile(_software_application(with_identifier=True))

    assert result["doi"] == "10.1234/example.1"
    assert result["identifiers"][0]["identifierType"] == "DOI"
    assert result["identifiers"][0]["identifier"] == "10.1234/example.1"
    assert (
        result["relatedIdentifiers"][0]["relatedIdentifier"]
        == "https://doi.org/10.1234/example.1"
    )

    assert result["creators"][0]["name"] == "Lovelace, Ada"
    assert (
        result["creators"][0]["nameIdentifiers"][0]["nameIdentifierScheme"] == "ORCID"
    )
    assert (
        result["creators"][0]["affiliation"][0]["affiliationIdentifierScheme"] == "ROR"
    )

    assert result["contributors"][0]["contributorType"] == "DataCurator"
    assert result["rights"][0]["rightsIdentifier"] == "Apache-2.0"


def test_datacite_transpile_generates_urn_when_identifier_missing(monkeypatch) -> None:
    monkeypatch.setattr("transpiler_mate.datacite.uuid.uuid4", lambda: "deadbeef")

    result = DataCiteTranspiler().transpile(
        _software_application(with_identifier=False)
    )

    assert result["identifiers"][0]["identifierType"] == "URN"
    assert result["identifiers"][0]["identifier"] == "urn:uuid:deadbeef"
    assert result["types"]["resourceTypeGeneral"] == "Software"


def test_datacite_transpile_omits_contributors_when_missing() -> None:
    app = _software_application(with_identifier=False)
    app.contributor = None

    result = DataCiteTranspiler().transpile(app)

    assert "contributors" not in result



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .metadata import MetadataManager, Transpiler
from .markdown import markdown_transpile
from datetime import datetime
from enum import auto, Enum
from functools import wraps
from loguru import logger
from pathlib import Path
from semver import Version
from typing import Tuple, Optional

import click
import json
import time


def _track(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()

        logger.info(
            f"Started at: {datetime.fromtimestamp(start_time).isoformat(timespec='milliseconds')}"
        )

        try:
            func(*args, **kwargs)

            logger.success(
                "------------------------------------------------------------------------"
            )
            logger.success("SUCCESS")
            logger.success(
                "------------------------------------------------------------------------"
            )
        except Exception as e:
            logger.error(
                "------------------------------------------------------------------------"
            )
            logger.error("FAIL")
            logger.error(e)
            logger.error(
                "------------------------------------------------------------------------"
            )

        end_time = time.time()

        logger.info(f"Total time: {end_time - start_time:.4f} seconds")
        logger.info(
            f"Finished at: {datetime.fromtimestamp(end_time).isoformat(timespec='milliseconds')}"
        )

    return wrapper


@click.group()
def main():
    pass


@main.command(context_settings={"show_default": True})
@click.argument(
    "source",
    type=click.Path(path_type=Path, exists=True, readable=True, resolve_path=True),
    required=True,
)
@click.option(
    "--base-url", type=click.STRING, required=True, help="The Invenio server base URL"
)
@click.option(
    "--auth-token",
    type=click.STRING,
    required=True,
    envvar="INVENIO_AUTH_TOKEN",
    help="The Invenio Access token",
)
@click.option(
    "--attach",
    type=click.Path(path_type=Path, exists=True, readable=True, resolve_path=True),
    multiple=True,
)
def invenio_publish(
    source: Path, base_url: str, auth_token: str, attach: Optional[Tuple[Path]]
):
    """
    Publishes the input CWL to an Invenio instance.
    """
    metadata_manager: MetadataManager = MetadataManager(source)

    logger.info(f"Interacting with Invenio server at {base_url})")

    from .invenio import InvenioMetadataTranspiler

    invenio_transpiler: InvenioMetadataTranspiler = InvenioMetadataTranspiler(
        metadata_manager=metadata_manager,
        invenio_base_url=base_url,
        auth_token=auth_token,
    )

    record_url = invenio_transpiler.create_or_update_process(
        source=source, attach=attach
    )

    logger.success(f"Record available on '{record_url}'")


def _transpile(source: Path, transpiler: Transpiler, output: Path):
    logger.info(f"Reading metadata from {source}...")
    metadata_manager: MetadataManager = MetadataManager(source)

    logger.success("Metadata successfully read!")
    logger.info("Transpiling metadata...")
    data = transpiler.transpile(metadata_manager.metadata)

    logger.success("Metadata successfully transpiled!")
    logger.info("Serializing metadata...")
    output.parent.mkdir(parents=True, exist_ok=True)
    with output.open("w") as output_stream:
        json.dump(data, output_stream, indent=2)

    logger.success(f"Metadata successfully serialized to {output}.")


@main.command(context_settings={"show_default": True})
@click.argument(
    "source",
    type=click.Path(path_type=Path, exists=True, readable=True, resolve_path=True),
    required=True,
)
@click.option(
    "--workflow-id", required=True, type=click.STRING, help="ID of the main Workflow"
)
@click.option(
    "--image-source",
    required=False,
    default=None,
    type=click.STRING,
    help="URL to get source code for building the image",
)
@click.option(
    "--image-revision",
    required=False,
    default=None,
    type=click.STRING,
    help="Source control revision identifier for the packaged software",
)
@click.option(
    "--output",
    type=click.Path(path_type=Path),
    required=False,
    default="annotations.json",
    help="The output file path",
)
def oci_annotations(
    source: Path,
    workflow_id: str,
    image_source: str | None,
    image_revision: str | None,
    output: Path,
):
    """
    Transpiles the input CWL to OCI annotations.
    """
    logger.info(f"Reading metadata from {source}...")
    metadata_manager: MetadataManager = MetadataManager(source)

    logger.success("Metadata successfully read!")
    logger.info("Transpiling metadata...")

    from .oci import OrasAnnotationsTranspiler
    from cwl_loader import load_cwl_from_yaml
    from cwl_loader.utils import search_process

    workflow = load_cwl_from_yaml(metadata_manager.raw_document)

    resolved_process = search_process(workflow_id, workflow)
    if not resolved_process:
        raise ValueError(
            f"Process {workflow_id} does not exist in input CWL document, only {list(map(lambda p: p.id, resolved_process)) if isinstance(resolved_process, list) else ['']} available."
        )

    data = OrasAnnotationsTranspiler(
        resolved_process, image_source, image_revision
    ).transpile(metadata_manager.metadata)

    logger.success("Metadata successfully transpiled!")
    logger.info("Serializing metadata...")
    output.parent.mkdir(parents=True, exist_ok=True)
    with output.open("w") as output_stream:
        json.dump(data, output_stream, indent=2)

    logger.success(f"Metadata successfully serialized to {output}.")


@main.command(context_settings={"show_default": True})
@click.argument(
    "source",
    type=click.Path(path_type=Path, exists=True, readable=True, resolve_path=True),
    required=True,
)
@click.option(
    "--code-repository",
    required=False,
    help="The (SVN, GitHub, CodePlex, ...) code repository URL",
)
@click.option(
    "--output",
    type=click.Path(path_type=Path),
    required=False,
    default="codemeta.json",
    help="The output file path",
)
def codemeta(source: Path, code_repository: str | None, output: Path):
    """
    Transpiles the input CWL to CodeMeta representation.
    """
    from .codemeta import CodeMetaTranspiler

    transpiler: CodeMetaTranspiler = CodeMetaTranspiler(code_repository)

    _transpile(source=source, transpiler=transpiler, output=output)


@main.command(context_settings={"show_default": True})
@click.argument(
    "source",
    type=click.Path(path_type=Path, exists=True, readable=True, resolve_path=True),
    required=True,
)
@click.option(
    "--output",
    type=click.Path(path_type=Path),
    required=False,
    default="record.json",
    help="The output file path",
)
def ogcrecord(source: Path, output: Path):
    """
    Transpiles the input CWL to OGC API Record.
    """
    from .ogcapi_records import OgcRecordsTranspiler

    transpiler = OgcRecordsTranspiler()

    _transpile(source=source, transpiler=transpiler, output=output)


@main.command(context_settings={"show_default": True})
@click.argument(
    "source",
    type=click.Path(path_type=Path, exists=True, readable=True, resolve_path=True),
    required=True,
)
@click.option(
    "--output",
    type=click.Path(path_type=Path),
    required=False,
    default="datacite.json",
    help="The output file path",
)
def datacite(source: Path, output: Path):
    """
    Transpiles the input CWL to DataCite Metadata.
    """
    from .datacite import DataCiteTranspiler

    transpiler = DataCiteTranspiler()

    _transpile(source=source, transpiler=transpiler, output=output)


@main.command(context_settings={"show_default": True})
@click.argument(
    "source",
    type=click.Path(path_type=Path, exists=True, readable=True, resolve_path=True),
    required=True,
)
@click.option(
    "--workflow-id",
    required=True,
    type=click.STRING,
    default="main",
    help="ID of the main Workflow",
)
@click.option(
    "--output",
    type=click.Path(path_type=Path),
    required=False,
    default=".",
    help="The output directory path",
)
@click.option(
    "--code-repository",
    required=False,
    help="The (SVN, GitHub, CodePlex, ...) code repository URL",
)
def markdown(source: Path, workflow_id: str, output: Path, code_repository: str | None):
    """
    Transpiles the input CWL to Markdown documentation.
    """
    output.mkdir(parents=True, exist_ok=True)

    target: Path = Path(output, f"{workflow_id}.md")

    logger.info(f"Rendering Markdown documentation of {source} to {target}...")

    with target.open("w") as output_stream:
        markdown_transpile(source, workflow_id, output_stream, code_repository)

    logger.info(f"Markdown documentation successfully serialized to {target}!")


class VersionPart(Enum):
    MAJOR = auto()
    MINOR = auto()
    PATCH = auto()
    BUILD = auto()
    PRE_RELEASE = auto()


@main.command(context_settings={"show_default": True})
@click.argument(
    "source",
    type=click.Path(path_type=Path, exists=True, readable=True, resolve_path=True),
    required=True,
)
@click.option(
    "--version-part",
    type=click.Choice(VersionPart, case_sensitive=False),
    required=False,
    default=VersionPart.MINOR,
    help="The version part to update",
)
def bump_version(source: Path, version_part: VersionPart):
    """
    Bumps the CWL SW version via SemVer Spec 2.0.0.
    """
    logger.info(f"Reading metadata from {source}...")
    metadata_manager: MetadataManager = MetadataManager(source)

    version = Version.parse(metadata_manager.metadata.software_version)

    if not version.is_valid:
        raise ValueError(
            f"Version {metadata_manager.metadata.software_version} is not compliant to the Semantic Versioning Specification 2.0.0, see https://semver.org/"
        )

    bumped_version = None

    match version_part:
        case VersionPart.MAJOR:
            bumped_version = version.bump_major()

        case VersionPart.MINOR:
            bumped_version = version.bump_minor()

        case VersionPart.PATCH:
            bumped_version = version.bump_patch()

        case VersionPart.BUILD:
            bumped_version = version.bump_build()

        case VersionPart.PRE_RELEASE:
            bumped_version = version.bump_prerelease()

        case _:
            raise ValueError(
                f"It's not you, it is us: {version_part} unsupported, but it shouldn't have been happened..."
            )

    logger.success(
        f"Software Version {metadata_manager.metadata.software_version} updated to {bumped_version}"
    )

    metadata_manager.metadata.software_version = str(bumped_version)

    metadata_manager.update()


for command in [
    bump_version,
    codemeta,
    datacite,
    invenio_publish,
    ogcrecord,
    oci_annotations,
]:
    command.callback = _track(command.callback)



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# generated by datamodel-codegen:
#   filename:  datacite_4_6.yaml
#   timestamp: 2026-02-12T10:56:31+00:00

from __future__ import annotations

from datetime import date as date_aliased
from enum import Enum
from typing import List, Optional, Union

from pydantic import AnyUrl, Field, RootModel
from transpiler_mate import TranspilerBaseModel


class Identifier(TranspilerBaseModel):
    """
    The Identifier is a unique string that identifies a resource.
    """

    identifier_type: str = Field(..., alias="identifierType")
    identifier: str


class NameIdentifier(TranspilerBaseModel):
    """
    Uniquely identifies an individual or legal entity, according to various schemes.
    """

    name_identifier: Optional[str] = Field(
        None,
        alias="nameIdentifier",
        description="Uniquely identifies an individual or legal entity, according to various schemes.",
    )
    name_identifier_scheme: str = Field(
        ...,
        alias="nameIdentifierScheme",
        description="The name of the name identifier scheme.",
    )
    scheme_uri: Optional[AnyUrl] = Field(
        None, alias="schemeURI", description="The URI of the name identifier scheme."
    )


class Affiliation(TranspilerBaseModel):
    """
    The organizational or institutional affiliation of the creator.
    """

    affiliation_identifier: Optional[str] = Field(
        None,
        alias="affiliationIdentifier",
        description="Uniquely identifies the organizational affiliation of the creator.",
    )
    affiliation_identifier_scheme: Optional[str] = Field(
        None,
        alias="affiliationIdentifierScheme",
        description="The name of the affiliation identifier scheme",
    )
    scheme_uri: Optional[AnyUrl] = Field(
        None,
        alias="schemeURI",
        description="The URI of the affiliation identifier scheme.",
    )


class NameType(Enum):
    """
    The type of name.
    """

    ORGANIZATIONAL = "Organizational"
    PERSONAL = "Personal"


class Creator(TranspilerBaseModel):
    """
    The main researcher involved in producing the data, or the author of the publication.
    """

    name: str = Field(..., description="The full name of the creator.")
    name_type: Optional[NameType] = Field(None, alias="nameType")
    given_name: Optional[str] = Field(
        None,
        alias="givenName",
        description="The personal or first name of the creator.",
    )
    family_name: Optional[str] = Field(
        None, alias="familyName", description="The surname or last name of the creator."
    )
    name_identifiers: Optional[List[NameIdentifier]] = Field(
        None,
        alias="nameIdentifiers",
        description="Uniquely identifies an individual or legal entity, according to various schemes.",
    )
    affiliation: Optional[List[Affiliation]] = Field(
        None,
        description="The organizational or institutional affiliations of the creator.",
    )


class TitleType(Enum):
    """
    The type of Title (other than the Main Title).
    """

    ALTERNATIVE_TITLE = "AlternativeTitle"
    SUBTITLE = "Subtitle"
    TRANSLATED_TITLE = "TranslatedTitle"
    OTHER = "Other"


class Title(TranspilerBaseModel):
    """
    A name or title by which a resource is known. May be the title of a dataset or the name of a piece of software or an instrument.
    """

    title: str = Field(..., description="A name or title by which a resource is known")
    lang: Optional[str] = Field(None, description="The languages of the title.")
    title_type: Optional[TitleType] = Field(
        None,
        alias="titleType",
        description="The type of Title (other than the Main Title).",
    )


class Publisher(TranspilerBaseModel):
    """
    The name of the entity that holds, archives, publishes, prints, distributes, releases, issues, or produces the resource. This property will be used to formulate the citation, so consider the prominence of the role.
    """

    name: str = Field(
        ...,
        description="The name of the entity that holds, archives, publishes, prints, distributes, releases, issues, or produces the resource. This property will be used to formulate the citation, so consider the prominence of the role.",
    )
    publisher_identifier: Optional[str] = Field(
        None,
        alias="publisherIdentifier",
        description="Uniquely identifies the publisher, according to various schemes.",
    )
    publisher_identifier_scheme: Optional[str] = Field(
        None,
        alias="publisherIdentifierScheme",
        description="The name of the publisher identifier scheme.",
    )
    scheme_uri: Optional[AnyUrl] = Field(
        None,
        alias="schemeURI",
        description="The URI of the publisher identifier scheme.",
    )
    lang: Optional[str] = Field(None, description="The language used by the Publisher.")


class Subject(TranspilerBaseModel):
    """
    Subject, keyword, classification code, or key phrase describing the resource.
    """

    subject: str = Field(
        ...,
        description="Subject, keyword, classification code, or key phrase describing the resource.",
    )
    subject_scheme: Optional[str] = Field(
        None,
        alias="subjectScheme",
        description="The name of the subject scheme or classification code or authority if one is used.",
    )
    scheme_uri: Optional[AnyUrl] = Field(
        None, alias="schemeURI", description="The URI of the subject identifier scheme."
    )
    value_uri: Optional[AnyUrl] = Field(
        None, alias="valueURI", description="The URI of the subject term."
    )
    classification_code: Optional[str] = Field(
        None,
        alias="classificationCode",
        description="The classification code used for the subject term in the subject schemes.",
    )
    lang: Optional[str] = Field(None, description="The language used in the Subject.")


class ContributorType(Enum):
    """
    The type of contributor of the resource
    """

    CONTACT_PERSON = "ContactPerson"
    DATA_COLLECTOR = "DataCollector"
    DATA_CURATOR = "DataCurator"
    DATA_MANAGER = "DataManager"
    DISTRIBUTOR = "Distributor"
    EDITOR = "Editor"
    HOSTING_INSTITUTION = "HostingInstitution"
    PRODUCER = "Producer"
    PROJECT_LEADER = "ProjectLeader"
    PROJECT_MANAGER = "ProjectManager"
    PROJECT_MEMBER = "ProjectMember"
    REGISTRATION_AGENCY = "RegistrationAgency"
    REGISTRATION_AUTHORITY = "RegistrationAuthority"
    RELATED_PERSON = "RelatedPerson"
    RESEARCHER = "Researcher"
    RESEARCH_GROUP = "ResearchGroup"
    RIGHTS_HOLDER = "RightsHolder"
    SPONSOR = "Sponsor"
    SUPERVISOR = "Supervisor"
    TRANSLATOR = "Translator"
    WORK_PACKAGE_LEADER = "WorkPackageLeader"
    OTHER = "Other"


class Contributor(Creator):
    """
    The institution or person responsible for collecting, managing, distributing, or otherwise contributing to the development of the resource.
    """

    contributor_type: ContributorType = Field(..., alias="contributorType")


class DateType(Enum):
    """
    The type of date
    """

    ACCEPTED = "Accepted"
    AVAILABLE = "Available"
    COPYRIGHTED = "Copyrighted"
    COLLECTED = "Collected"
    COVERAGE = "Coverage"
    CREATED = "Created"
    ISSUED = "Issued"
    SUBMITTED = "Submitted"
    UPDATED = "Updated"
    VALID = "Valid"
    WITHDRAWN = "Withdrawn"
    OTHER = "Other"


class Date(TranspilerBaseModel):
    """
    Date relevant to the work.
    """

    date: date_aliased = Field(..., description="Date relevant to the work.")
    date_type: DateType = Field(..., alias="dateType", description="The type of date")
    date_information: Optional[str] = Field(
        None,
        alias="dateInformation",
        description="Specific information about the date, if appropriate.",
    )


class ResourceTypeGeneral(Enum):
    """
    The general type of a resource.
    """

    AUDIOVISUAL = "Audiovisual"
    AWARD = "Award"
    BOOK = "Book"
    BOOK_CHAPTER = "BookChapter"
    COLLECTION = "Collection"
    COMPUTATIONAL_NOTEBOOK = "ComputationalNotebook"
    CONFERENCE_PAPER = "ConferencePaper"
    CONFERENCE_PROCEEDING = "ConferenceProceeding"
    DATA_PAPER = "DataPaper"
    DATASET = "Dataset"
    DISSERTATION = "Dissertation"
    EVENT = "Event"
    IMAGE = "Image"
    INTERACTIVE_RESOURCE = "InteractiveResource"
    INSTRUMENT = "Instrument"
    JOURNAL = "Journal"
    JOURNAL_ARTICLE = "JournalArticle"
    MODEL = "Model"
    OUTPUT_MANAGEMENT_PLAN = "OutputManagementPlan"
    PEER_REVIEW = "PeerReview"
    PHYSICAL_OBJECT = "PhysicalObject"
    PREPRINT = "Preprint"
    PROJECT = "Project"
    REPORT = "Report"
    SERVICE = "Service"
    SOFTWARE = "Software"
    SOUND = "Sound"
    STANDARD = "Standard"
    STUDY_REGISTRATION = "StudyRegistration"
    TEXT = "Text"
    WORKFLOW = "Workflow"
    OTHER = "Other"


class ResourceType(TranspilerBaseModel):
    """
    A description of the resource.
    """

    resource_type: str = Field(
        ..., alias="resourceType", description="A description of the resource."
    )
    resource_type_general: ResourceTypeGeneral = Field(..., alias="resourceTypeGeneral")


class AlternateIdentifier(TranspilerBaseModel):
    """
    An identifier other than the primary Identifier applied to the resource being registered.
    """

    alternate_identifier: Optional[str] = Field(
        None,
        alias="alternateIdentifier",
        description="An identifier other than the primary Identifier applied to the resource being registered",
    )
    alternate_identifier_type: ResourceTypeGeneral = Field(
        ..., alias="alternateIdentifierType"
    )


class RelationType(Enum):
    """
    Description of the relationship of the resource being registered (A) and the related resource (B).
    """

    IS_CITED_BY = "IsCitedBy"
    CITES = "Cites"
    IS_SUPPLEMENT_TO = "IsSupplementTo"
    IS_SUPPLEMENTED_BY = "IsSupplementedBy"
    IS_CONTINUED_BY = "IsContinuedBy"
    CONTINUES = "Continues"
    IS_DESCRIBED_BY = "IsDescribedBy"
    DESCRIBES = "Describes"
    HAS_METADATA = "HasMetadata"
    IS_METADATA_FOR = "IsMetadataFor"
    HAS_VERSION = "HasVersion"
    IS_VERSION_OF = "IsVersionOf"
    IS_NEW_VERSION_OF = "IsNewVersionOf"
    IS_PREVIOUS_VERSION_OF = "IsPreviousVersionOf"
    IS_PART_OF = "IsPartOf"
    HAS_PART = "HasPart"
    IS_PUBLISHED_IN = "IsPublishedIn"
    IS_REFERENCED_BY = "IsReferencedBy"
    REFERENCES = "References"
    IS_DOCUMENTED_BY = "IsDocumentedBy"
    DOCUMENTS = "Documents"
    IS_COMPILED_BY = "IsCompiledBy"
    COMPILES = "Compiles"
    IS_VARIANT_FORM_OF = "IsVariantFormOf"
    IS_ORIGINAL_FORM_OF = "IsOriginalFormOf"
    IS_IDENTICAL_TO = "IsIdenticalTo"
    IS_REVIEWED_BY = "IsReviewedBy"
    REVIEWS = "Reviews"
    IS_DERIVED_FROM = "IsDerivedFrom"
    IS_SOURCE_OF = "IsSourceOf"
    IS_REQUIRED_BY = "IsRequiredBy"
    REQUIRES = "Requires"
    IS_OBSOLETED_BY = "IsObsoletedBy"
    OBSOLETES = "Obsoletes"
    IS_COLLECTED_BY = "IsCollectedBy"
    COLLECTS = "Collects"
    IS_TRANSLATION_OF = "IsTranslationOf"
    HAS_TRANSLATION = "HasTranslation"


class RelatedIdentifierType(Enum):
    """
    The type of the RelatedIdentifier.
    """

    ARK = "ARK"
    AR_XIV = "arXiv"
    BIBCODE = "bibcode"
    CSTR = "CSTR"
    DOI = "DOI"
    EAN13 = "EAN13"
    EISSN = "EISSN"
    HANDLE = "Handle"
    IGSN = "IGSN"
    ISBN = "ISBN"
    ISSN = "ISSN"
    ISTC = "ISTC"
    LISSN = "LISSN"
    LSID = "LSID"
    PMID = "PMID"
    PURL = "PURL"
    RRID = "RRID"
    UPC = "UPC"
    URL = "URL"
    URN = "URN"
    W3ID = "w3id"


class RelatedIdentifier(TranspilerBaseModel):
    """
    Identifier of related resources.
    """

    related_identifier: Optional[str] = Field(
        None, alias="relatedIdentifier", description="Identifier of related resources."
    )
    related_identifier_type: Optional[RelatedIdentifierType] = Field(
        None, alias="relatedIdentifierType"
    )
    relation_type: Optional[RelationType] = Field(None, alias="relationType")
    related_metadata_scheme: Optional[str] = Field(
        None, alias="relatedMetadataScheme", description="The name of the schemes."
    )
    scheme_uri: Optional[AnyUrl] = Field(
        None, alias="schemeURI", description="The URI of the name identifier scheme."
    )
    scheme_type: Optional[str] = Field(
        None,
        alias="schemeType",
        description="The type of the relatedMetadataScheme, linked with the schemeURI",
    )
    resource_type_general: Optional[ResourceTypeGeneral] = Field(
        None, alias="resourceTypeGeneral"
    )


class Right(TranspilerBaseModel):
    """
    Any right information for this resource.
    """

    rights: str = Field(..., description="Any right information for this resource.")
    rights_uri: Optional[AnyUrl] = Field(
        None, alias="rightsURI", description="The URI of the license."
    )
    rights_identifier: Optional[str] = Field(
        None,
        alias="rightsIdentifier",
        description="A short, standardized version of the license name.",
    )
    rights_identifier_scheme: Optional[str] = Field(
        None, alias="rightsIdentifierScheme", description="The name of the scheme."
    )
    scheme_uri: Optional[AnyUrl] = Field(
        None, alias="schemeURI", description="The URI of the rightsIdentifierScheme."
    )


class DescriptionType(Enum):
    """
    The type of the Description.
    """

    ABSTRACT = "Abstract"
    METHODS = "Methods"
    SERIES_INFORMATION = "SeriesInformation"
    TABLE_OF_CONTENTS = "TableOfContents"
    TECHNICAL_INFO = "TechnicalInfo"
    OTHER = "Other"


class Description(TranspilerBaseModel):
    """
    All additional information that does not fit in any of the other categories.
    """

    description: str = Field(
        ...,
        description="All additional information that does not fit in any of the other categories.",
    )
    description_type: DescriptionType = Field(
        ..., alias="descriptionType", description="The type of the Description."
    )


class GeoLocationPoint(TranspilerBaseModel):
    """
    A point location in space.
    """

    point_longitude: float = Field(
        ..., alias="pointLongitude", description="Longitudinal dimension of point."
    )
    point_latitude: float = Field(
        ..., alias="pointLatitude", description="Latitudinal dimension of point."
    )


class GeoLocationBox(TranspilerBaseModel):
    """
    The spatial limits of a box.
    """

    west_bound_longitude: float = Field(
        ...,
        alias="westBoundLongitude",
        description="Western longitudinal dimension of box.",
    )
    east_bound_longitude: float = Field(
        ...,
        alias="eastBoundLongitude",
        description="Eastern longitudinal dimension of box.",
    )
    south_bound_latitude: float = Field(
        ...,
        alias="southBoundLatitude",
        description="Southern latitudinal dimension of box.",
    )
    north_bound_latitude: float = Field(
        ...,
        alias="northBoundLatitude",
        description="Northern latitudinal dimension of box.",
    )


class GeoLocationPolygon(TranspilerBaseModel):
    """
    A drawn polygon area, defined by a set of points and lines connecting the points in a closed chain.
    """

    polygon_point: Optional[GeoLocationPoint] = Field(None, alias="polygonPoint")
    in_polygon_point: Optional[GeoLocationPoint] = Field(None, alias="inPolygonPoint")


class GeoLocation(TranspilerBaseModel):
    """
    Spatial region or named place where the data was gathered or about which the data is focused.
    """

    geo_location_point: Optional[GeoLocationPoint] = Field(
        None, alias="geoLocationPoint"
    )
    geo_location_box: Optional[GeoLocationBox] = Field(None, alias="geoLocationBox")
    geo_location_place: Optional[str] = Field(
        None,
        alias="geoLocationPlace",
        description="Description of a geographic location.",
    )
    geo_location_polygon: Optional[List[GeoLocationPolygon]] = Field(
        None, alias="geoLocationPolygon", min_length=4
    )


class FunderIdentifierType(Enum):
    """
    The type of the funderIdentifier.
    """

    CROSSREF_FUNDER_ID = "Crossref Funder ID"
    GRID = "GRID"
    ISNI = "ISNI"
    ROR = "ROR"
    OTHER = "Other"


class FundingReference(TranspilerBaseModel):
    """
    Information about financial support (funding) for the resource being registered.
    """

    funder_name: str = Field(
        ..., alias="funderName", description="Name of the funding provider."
    )
    funder_identifier: Optional[str] = Field(
        None,
        alias="funderIdentifier",
        description="Uniquely identifies a funding entity, according to various types",
    )
    funder_identifier_type: FunderIdentifierType = Field(
        ...,
        alias="funderIdentifierType",
        description="The type of the funderIdentifier.",
    )
    scheme_uri: Optional[AnyUrl] = Field(
        None, alias="schemeURI", description="The URI of the funder identifier scheme."
    )
    award_number: Optional[str] = Field(
        None,
        alias="awardNumber",
        description="The code assigned by the funder to a sponsored award (grant).",
    )
    award_uri: Optional[AnyUrl] = Field(
        None,
        alias="awardURI",
        description="The URI leading to a page provided by the funder for more information about the award (grant).",
    )
    award_title: Optional[str] = Field(
        None,
        alias="awardTitle",
        description="The human readable title or name of the award (grant).",
    )


class RelatedItemIdentifier(TranspilerBaseModel):
    """
    The identifier for the related item.
    """

    related_item_identifier_type: Optional[RelatedIdentifierType] = Field(
        None, alias="relatedItemIdentifierType"
    )
    related_metadata_scheme: Optional[str] = Field(
        None, alias="relatedMetadataScheme", description="The name of the schemes."
    )
    scheme_uri: Optional[AnyUrl] = Field(
        None, alias="schemeURI", description="The URI of the name identifier scheme."
    )
    scheme_type: Optional[str] = Field(
        None,
        alias="schemeType",
        description="The type of the relatedMetadataScheme, linked with the schemeURI",
    )


class RelatedItemCreator(TranspilerBaseModel):
    """
    The institution or person responsible for creating the related resource.
    """

    name: str = Field(..., description="The full name of the related item creator")
    name_type: Optional[NameType] = Field(None, alias="nameType")
    given_name: Optional[str] = Field(
        None,
        alias="givenName",
        description="The personal or first name of the creator.",
    )
    family_name: Optional[str] = Field(
        None, alias="familyName", description="The surname or last name of the creator."
    )


class RelatedItemTitle(TranspilerBaseModel):
    """
    Title of the related item.
    """

    title: str = Field(..., description="Title of the related item.")
    title_type: Optional[str] = Field(
        None, alias="titleType", description="Type of the related item title."
    )


class RelatedItemContributor(RelatedItemCreator):
    """
    An institution or person identified as contributing to the development of the resource.
    """

    contributor_type: ContributorType = Field(..., alias="contributorType")


class NumberType(Enum):
    """
    Type of the related item’s number, e.g., report number or article number.
    """

    ARTICLE = "Article"
    CHAPTER = "Chapter"
    REPORT = "Report"
    OTHER = "Other"


class PublicationYear1(RootModel[str]):
    root: str = Field(
        ...,
        description="The year when the data was or will be made publicly available.",
        pattern="^\\d{4}$",
    )


class Event(Enum):
    """
    Indicates a state-change action for the DOI
    """

    PUBLISH = "publish"
    REGISTER = "register"
    HIDE = "hide"


class RelatedItem(TranspilerBaseModel):
    """
    Information about a resource related to the one being registered.
    """

    related_item_type: ResourceTypeGeneral = Field(..., alias="relatedItemType")
    relation_type: ResourceTypeGeneral = Field(..., alias="relationType")
    related_item_identifier: Optional[RelatedItemIdentifier] = Field(
        None, alias="relatedItemIdentifier"
    )
    creators: Optional[List[RelatedItemCreator]] = Field(
        None,
        description="The institution or person responsible for creating the related resource.",
    )
    titles: List[RelatedItemTitle] = Field(
        ..., description="Title of the related item", min_length=1
    )
    publication_year: Optional[Union[int, PublicationYear1]] = Field(
        None,
        alias="publicationYear",
        description="The year when the data was or will be made publicly available.",
    )
    volume: Optional[str] = Field(None, description="Volume of the related item.")
    issue: Optional[str] = Field(
        None, description="Issue number or name of the related item."
    )
    number: Optional[str] = Field(
        None,
        description="Number of the resource within the related item, e.g., report number or article number.",
    )
    number_type: Optional[NumberType] = Field(
        None,
        alias="numberType",
        description="Type of the related item’s number, e.g., report number or article number.",
    )
    first_page: Optional[str] = Field(
        None,
        alias="firstPage",
        description="First page of the resource within the related item, e.g., of the chapter, article, or conference paper in proceedings.",
    )
    last_page: Optional[str] = Field(
        None,
        alias="lastPage",
        description="Last page of the resource within the related item, e.g., of the chapter, article, or conference paper in proceedings.",
    )
    publisher: Optional[str] = Field(
        None,
        description="The name of the entity that holds, archives, publishes prints, distributes, releases, issues, or produces the resource.",
    )
    edition: Optional[str] = Field(
        None, description="Edition or version of the related item."
    )
    contributors: Optional[List[RelatedItemContributor]] = Field(
        None,
        description="An institution or person identified as contributing to the development of the resource",
    )


class DataCiteAttributes(TranspilerBaseModel):
    """
    DataCite Metadata Schema
    """

    doi: Optional[str] = Field(None, description="The full DOI (prefix + suffix)")
    prefix: Optional[str] = Field(None, description="The namespace prefix")
    suffix: Optional[str] = Field(None, description="The suffix portion of the DOI")
    event: Optional[Event] = Field(
        None, description="Indicates a state-change action for the DOI"
    )
    identifiers: List[Identifier]
    creators: List[Creator] = Field(
        ...,
        description="The main researchers involved in producing the data, or the authors of the publication, in priority order.",
        min_length=1,
    )
    titles: List[Title] = Field(
        ...,
        description="Names or titles by which a resource is known. May be the title of a dataset or the name of a piece of software or an instrument.",
        min_length=1,
    )
    publisher: Publisher
    publication_year: Union[int, PublicationYear1] = Field(
        ...,
        alias="publicationYear",
        description="The year when the data was or will be made publicly available.",
    )
    subjects: Optional[List[Subject]] = Field(
        None,
        description="Subjects, keywords, classification codes, or key phrases describing the resource.",
    )
    contributors: Optional[List[Contributor]] = Field(
        None,
        description="The institution or person responsible for collecting, managing, distributing, or otherwise contributing to the development of the resource.",
    )
    dates: Optional[List[Date]] = Field(
        None, description="Different dates relevant to the work."
    )
    language: Optional[str] = Field(
        None, description="The primary language of the resource"
    )
    types: ResourceType
    alternate_identifiers: Optional[List[AlternateIdentifier]] = Field(
        None,
        alias="alternateIdentifiers",
        description="An identifier other than the primary Identifier applied to the resource being registered.",
    )
    related_identifiers: Optional[List[RelatedIdentifier]] = Field(
        None,
        alias="relatedIdentifiers",
        description="Identifiers of related resources.",
    )
    sizes: Optional[List[str]] = Field(
        None,
        description="Size (e.g., bytes, pages, inches, etc.) or duration (extent), e.g., hours, minutes, days, etc., of a resource.",
    )
    formats: Optional[List[str]] = Field(
        None, description="Technical format of the resources."
    )
    version: Optional[str] = Field(
        None, description="The version number of the resources."
    )
    rights_list: Optional[List[Right]] = Field(
        None, alias="rightsList", description="Any rights information for this resource"
    )
    descriptions: Optional[List[Description]] = None
    geo_locations: Optional[List[GeoLocation]] = Field(
        None,
        alias="geoLocations",
        description="Spatial regions or named places where the data was gathered or about which the data is focused.",
    )
    funding_references: Optional[List[FundingReference]] = Field(
        None,
        alias="fundingReferences",
        description="Information about financial support (funding) for the resource being registered.",
    )
    related_items: Optional[List[RelatedItem]] = Field(
        None,
        alias="relatedItems",
        description="Informations about a resource related to the one being registered.",
    )


class Data(TranspilerBaseModel):
    """
    TODO
    """

    id: str
    type: str
    attributes: Optional[DataCiteAttributes] = None


class DataCiteMetadata46(TranspilerBaseModel):
    """
    DataCite Metadata Schema
    """

    data: Data


class Model(RootModel[DataCiteMetadata46]):
    root: DataCiteMetadata46



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from cwl_loader import load_cwl_from_yaml
from cwl_loader.utils import to_index
from cwl_loader.utils import search_process
from cwl_utils.parser import Process, cwl_v1_0, cwl_v1_1, cwl_v1_2
from datetime import datetime
from importlib.metadata import version, PackageNotFoundError
from jinja2 import Environment, PackageLoader
from loguru import logger
from pathlib import Path
from transpiler_mate.metadata import MetadataManager
from transpiler_mate.codemeta import CodeMetaTranspiler
from typing import (
    get_args,
    get_origin,
    Any,
    List,
    Literal,
    Mapping,
    TextIO,
    Union,
)

import time

# START custom built-in functions to simplify the CWL rendering

# CWLtype to string methods

NA_ROLE = "N/A"

PERSON_TYPE = "https://schema.org/Person"
ROLE_TYPE = "https://schema.org/Role"

RoleKind = Literal["author", "contributor"]

InputRecordSchema = Union[
    cwl_v1_0.InputRecordSchema, cwl_v1_1.InputRecordSchema, cwl_v1_2.InputRecordSchema
]

SchemaDefRequirement = Union[
    cwl_v1_0.SchemaDefRequirement,
    cwl_v1_1.SchemaDefRequirement,
    cwl_v1_2.SchemaDefRequirement,
]


def _normalize_role_people(value: Any, *, person_key: RoleKind) -> list[dict]:
    """
    Normalize a JSON-ish field that can be:
      - None
      - dict representing Person
      - dict representing Role (schema.org Role) holding a person under `person_key`
      - list (possibly nested) of any of the above

    into:
      - list of dicts, each a Role-shaped dict with:
          {"@type": "https://schema.org/Role", "role_name": <str>, person_key: <Person dict>, ...}

    For a bare Person dict, wraps it into a Role dict with role_name = "N/A".
    """

    def looks_like_person(d: dict) -> bool:
        return d.get("@type") == PERSON_TYPE or any(
            k in d
            for k in ("given_name", "family_name", "email", "identifier", "affiliation")
        )

    def ensure_person_dict(d: dict) -> dict:
        # shallow copy; ensure @type if you rely on it
        p = dict(d)
        p.setdefault("@type", PERSON_TYPE)
        return p

    def ensure_role_dict(role: dict) -> dict:
        r = dict(role)  # avoid mutating caller
        r.setdefault("@type", ROLE_TYPE)
        r.setdefault("role_name", NA_ROLE)
        # Normalize the nested person payload too (if present and dict)
        if isinstance(r.get(person_key), dict):
            r[person_key] = ensure_person_dict(r[person_key])
        return r

    if value is None:
        return []

    if isinstance(value, list):
        out: list[dict] = []
        for item in value:
            out.extend(_normalize_role_people(item, person_key=person_key))
        return out

    if not isinstance(value, dict):
        raise TypeError(f"{person_key} must be dict/list/None, got {type(value)!r}")

    # Already a Role wrapper if it contains the expected person key
    if person_key in value:
        return [ensure_role_dict(value)]

    # Otherwise treat it as a Person and wrap
    if looks_like_person(value):
        person = ensure_person_dict(value)
        return [ensure_role_dict({person_key: person})]

    raise ValueError(
        f"Unrecognized {person_key} dict shape (keys={sorted(value.keys())})"
    )


def normalize_author(value: Any) -> list[dict]:
    return _normalize_role_people(value, person_key="author")


def normalize_contributor(value: Any) -> list[dict]:
    return _normalize_role_people(value, person_key="contributor")


def type_to_string(typ: Any, parent: Process) -> str:
    """
    Serializes a CWL type to a human-readable string.

    Args:
        `typ` (`Any`): Any CWL type

    Returns:
        `str`: The human-readable string representing the input CWL type.
    """
    if get_origin(typ) is Union:
        return f"One of:<ul>{''.join(f'<li>{type_to_string(inner_type, parent)}</li>' for inner_type in get_args(typ))}</ul>"

    if isinstance(typ, list):
        return f"One of:<ul>{''.join(f'<li>{type_to_string(t, parent)}</li>' for t in typ)}</ul>"

    if hasattr(typ, "items"):
        return f"`array` of {type_to_string(typ.items, parent)}"

    if isinstance(typ, InputRecordSchema):
        fields = (
            "".join(
                f"<li>`{field.name.split('/')[-1]}`: {type_to_string(field.type_, parent)}</li>"
                for field in typ.fields
            )
            if typ.fields
            else ""
        )

        return f"[{typ.name.split('#')[-1]}]({typ.name}):<ul>{fields}</ul>"

    if isinstance(typ, str):
        type_str = typ
    elif hasattr(typ, "__name__"):
        type_str = typ.__name__
    elif hasattr(typ, "type_"):
        type_str = typ.type_
    else:
        # last hope to follow back
        type_str = str(typ)

    if "#" in type_str:  # we can assume it is an URL
        if parent.requirements:
            for requirement in parent.requirements:
                if isinstance(requirement, SchemaDefRequirement):
                    for inner_type in requirement.types:
                        if type_str == inner_type.name:
                            return type_to_string(inner_type, parent)

        # follow up on plain link if not found
        return f"[{type_str.split('#')[-1]}]({type_str})"

    for special_type in ["Any", "Directory", "File"]:
        if special_type == type_str:
            return (
                f"[{type_str}](https://www.commonwl.org/v1.2/Workflow.html#{type_str})"
            )

    if "enum" == type_str:
        symbols = "".join(
            f"<li>`{symbol.split('/')[-1]}`</li>"
            for symbol in typ.symbols  # type: ignore
        )
        return f"[{type_str}](https://www.commonwl.org/v1.2/Workflow.html#{type(typ).__name__}):<ul>{symbols}</ul>"

    return f"[{type_str}](https://www.commonwl.org/v1.2/Workflow.html#CWLType)"


def _get_version() -> str:
    try:
        return version("transpiler_mate")
    except PackageNotFoundError:
        return "N/A"


def _to_mapping(functions: List[Any]) -> Mapping[str, Any]:
    mapping: Mapping[str, Any] = {}

    for function in functions:
        mapping[function.__name__] = function

    return mapping


def nullable(type_: Any) -> bool:
    return (
        isinstance(type_, list)
        and "null" in type_
        or hasattr(type_, "items")
        and nullable(getattr(type_, "items"))
    )


def get_exection_command(clt: Any) -> str:
    result: List[str] = []

    def _append_arg(arg: Any):
        if isinstance(arg, list):
            for arg_i in arg:
                _append_arg(arg_i)
        elif isinstance(arg, str):
            result.append(arg)
        else:
            result.append("<ARGUMENT_DYNAMICALLY_SET>")

    def _check_then_append(arg_name: str):
        if hasattr(clt, arg_name) and getattr(clt, arg_name):
            _append_arg(getattr(clt, arg_name))

    _check_then_append("baseCommand")
    _check_then_append("arguments")

    return " ".join(result)


_jinja_environment = Environment(
    loader=PackageLoader(package_name="transpiler_mate.markdown")
)
_jinja_environment.globals["type_to_string"] = type_to_string
_jinja_environment.filters.update(
    _to_mapping(
        [
            get_exection_command,
            normalize_author,
            normalize_contributor,
        ]
    )
)
_jinja_environment.tests.update(_to_mapping([nullable]))

# END


def markdown_transpile(
    source: Path, workflow_id: str, output_stream: TextIO, code_repository: str | None
):
    logger.info(f"Reading metadata from {source}...")
    metadata_manager: MetadataManager = MetadataManager(source)

    logger.success("Metadata successfully read!")
    logger.info("Transpiling metadata...")

    transpiler: CodeMetaTranspiler = CodeMetaTranspiler(code_repository)
    metadata = transpiler.transpile(metadata_manager.metadata)

    logger.success("Metadata successfully transpiled!")
    logger.info("Reading Workflow model...")

    cwl_document = load_cwl_from_yaml(metadata_manager.raw_document)

    process = search_process(workflow_id, cwl_document)
    if not process:
        available = (
            list(map(lambda p: getattr(p, "id", str(p)), cwl_document))
            if isinstance(cwl_document, list)
            else [getattr(cwl_document, "id", str(cwl_document))]
        )
        raise ValueError(
            f"Workflow {workflow_id} does not exist in input CWL document, only {available} available."
        )

    logger.success("Workflow model successfully read!")

    template = _jinja_environment.get_template("index.md")

    output_stream.write(
        template.render(
            version=_get_version(),
            timestamp=datetime.fromtimestamp(time.time()).isoformat(
                timespec="milliseconds"
            ),
            software_source_code=metadata
            if "SoftwareSourceCode" == metadata["@type"]
            else None,
            software_application=metadata["targetProduct"]
            if "SoftwareSourceCode" == metadata["@type"]
            else metadata,
            workflow=process,
            index=to_index(cwl_document)
            if isinstance(cwl_document, list)
            else {workflow_id: cwl_document},
        )
    )



# Copyright 2026 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# generated by datamodel-codegen:
#   filename:  oci_annotations.yaml
#   timestamp: 2026-03-10T11:35:32+00:00

from __future__ import annotations

from typing import Optional

from pydantic import Field, RootModel
from transpiler_mate import TranspilerBaseModel


class OciAnnotations(TranspilerBaseModel):
    org_opencontainers_image_title: Optional[str] = Field(
        None, alias="org.opencontainers.image.title"
    )
    org_opencontainers_image_description: Optional[str] = Field(
        None, alias="org.opencontainers.image.description"
    )
    org_opencontainers_image_version: Optional[str] = Field(
        None, alias="org.opencontainers.image.version"
    )
    org_opencontainers_image_source: Optional[str] = Field(
        None, alias="org.opencontainers.image.source"
    )
    org_opencontainers_image_revision: Optional[str] = Field(
        None, alias="org.opencontainers.image.revision"
    )
    org_opencontainers_image_licenses: Optional[str] = Field(
        None, alias="org.opencontainers.image.licenses"
    )
    org_cwl_entrypoint: Optional[str] = Field(None, alias="org.cwl.entrypoint")
    org_cwl_spec: Optional[str] = Field(None, alias="org.cwl.spec")
    org_cwl_type: Optional[str] = Field(None, alias="org.cwl.type")


class Model(RootModel[OciAnnotations]):
    root: OciAnnotations



# Copyright 2026 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import sys
from pathlib import Path
from types import SimpleNamespace

from transpiler_mate import cli


class DummyTranspiler:
    def transpile(self, metadata_source):
        return {"name": metadata_source.name, "ok": True}


class DummyManager:
    def __init__(self, source: Path):
        self.document_source = source
        self.metadata = SimpleNamespace(name="Example", software_version="1.2.3")
        self.updated = False

    def update(self) -> None:
        self.updated = True


class DummyManagerFactory:
    def __init__(self):
        self.instance = None

    def __call__(self, source: Path):
        self.instance = DummyManager(source)
        return self.instance


def test_transpile_writes_json_output(monkeypatch, tmp_path: Path) -> None:
    source = tmp_path / "input.cwl"
    source.write_text("cwlVersion: v1.2\n", encoding="utf-8")

    factory = DummyManagerFactory()
    monkeypatch.setattr(cli, "MetadataManager", factory)

    output = tmp_path / "out" / "result.json"
    cli._transpile(source=source, transpiler=DummyTranspiler(), output=output)

    assert output.exists()
    assert json.loads(output.read_text(encoding="utf-8")) == {
        "name": "Example",
        "ok": True,
    }


def test_track_catches_exceptions() -> None:
    calls = {"count": 0}

    @cli._track
    def failing():
        calls["count"] += 1
        raise RuntimeError("boom")

    # The wrapper should not re-raise.
    failing()

    assert calls["count"] == 1


def test_bump_version_updates_metadata(monkeypatch, tmp_path: Path) -> None:
    source = tmp_path / "workflow.cwl"
    source.write_text("cwlVersion: v1.2\n", encoding="utf-8")

    factory = DummyManagerFactory()
    monkeypatch.setattr(cli, "MetadataManager", factory)

    # Call undecorated callback to avoid _track swallowing potential failures.
    cli.bump_version.callback.__wrapped__(
        source=source, version_part=cli.VersionPart.PATCH
    )

    assert factory.instance is not None
    assert factory.instance.metadata.software_version == "1.2.4"
    assert factory.instance.updated is True


def test_bump_version_raises_for_invalid_semver(monkeypatch, tmp_path: Path) -> None:
    source = tmp_path / "workflow.cwl"
    source.write_text("cwlVersion: v1.2\n", encoding="utf-8")

    manager = DummyManager(source)
    manager.metadata.software_version = "not-a-version"
    monkeypatch.setattr(cli, "MetadataManager", lambda _source: manager)

    try:
        cli.bump_version.callback.__wrapped__(
            source=source, version_part=cli.VersionPart.PATCH
        )
        assert False, "Expected ValueError"
    except ValueError:
        assert True


def test_codemeta_command_uses_transpile_with_codemeta_transpiler(
    monkeypatch, tmp_path: Path
) -> None:
    source = tmp_path / "workflow.cwl"
    output = tmp_path / "codemeta.json"
    source.write_text("cwlVersion: v1.2\n", encoding="utf-8")

    captured = {}

    class DummyCodeMetaTranspiler:
        def __init__(self, code_repository):
            self.code_repository = code_repository

    monkeypatch.setitem(
        sys.modules,
        "transpiler_mate.codemeta",
        SimpleNamespace(CodeMetaTranspiler=DummyCodeMetaTranspiler),
    )
    monkeypatch.setattr(
        cli,
        "_transpile",
        lambda source, transpiler, output: captured.update(
            {"source": source, "transpiler": transpiler, "output": output}
        ),
    )

    cli.codemeta.callback.__wrapped__(
        source=source,
        code_repository="https://github.com/acme/example-tool",
        output=output,
    )

    assert captured["source"] == source
    assert captured["output"] == output
    assert (
        captured["transpiler"].code_repository == "https://github.com/acme/example-tool"
    )


def test_datacite_and_ogcrecord_commands_delegate_to_transpile(
    monkeypatch, tmp_path: Path
) -> None:
    source = tmp_path / "workflow.cwl"
    source.write_text("cwlVersion: v1.2\n", encoding="utf-8")

    calls = []
    monkeypatch.setitem(
        sys.modules,
        "transpiler_mate.datacite",
        SimpleNamespace(DataCiteTranspiler=lambda: "datacite-transpiler"),
    )
    monkeypatch.setitem(
        sys.modules,
        "transpiler_mate.ogcapi_records",
        SimpleNamespace(OgcRecordsTranspiler=lambda: "ogc-transpiler"),
    )
    monkeypatch.setattr(
        cli,
        "_transpile",
        lambda source, transpiler, output: calls.append((source, transpiler, output)),
    )

    cli.datacite.callback.__wrapped__(source=source, output=tmp_path / "datacite.json")
    cli.ogcrecord.callback.__wrapped__(source=source, output=tmp_path / "record.json")

    assert calls[0][1] == "datacite-transpiler"
    assert calls[1][1] == "ogc-transpiler"


def test_oci_annotations_command_writes_output(monkeypatch, tmp_path: Path) -> None:
    source = tmp_path / "workflow.cwl"
    source.write_text("cwlVersion: v1.2\n", encoding="utf-8")
    output = tmp_path / "annotations.json"

    metadata_manager = SimpleNamespace(
        raw_document={"id": "#main"},
        metadata=SimpleNamespace(name="Example"),
    )
    monkeypatch.setattr(cli, "MetadataManager", lambda _source: metadata_manager)
    monkeypatch.setitem(
        sys.modules,
        "cwl_loader",
        SimpleNamespace(load_cwl_from_yaml=lambda _raw: "doc"),
    )
    monkeypatch.setitem(
        sys.modules,
        "cwl_loader.utils",
        SimpleNamespace(
            search_process=lambda workflow_id, workflow: SimpleNamespace(id="#main")
        ),
    )
    monkeypatch.setitem(
        sys.modules,
        "transpiler_mate.oci",
        SimpleNamespace(
            OrasAnnotationsTranspiler=lambda process, image_source, image_revision: (
                SimpleNamespace(transpile=lambda _metadata: {"ok": True})
            )
        ),
    )

    cli.oci_annotations.callback.__wrapped__(
        source=source,
        workflow_id="main",
        image_source=None,
        image_revision=None,
        output=output,
    )

    assert json.loads(output.read_text(encoding="utf-8")) == {"ok": True}


def test_oci_annotations_command_raises_when_process_missing(
    monkeypatch, tmp_path: Path
) -> None:
    source = tmp_path / "workflow.cwl"
    source.write_text("cwlVersion: v1.2\n", encoding="utf-8")

    metadata_manager = SimpleNamespace(
        raw_document={"id": "#main"},
        metadata=SimpleNamespace(name="Example"),
    )
    monkeypatch.setattr(cli, "MetadataManager", lambda _source: metadata_manager)
    monkeypatch.setitem(
        sys.modules,
        "cwl_loader",
        SimpleNamespace(load_cwl_from_yaml=lambda _raw: "doc"),
    )
    monkeypatch.setitem(
        sys.modules,
        "cwl_loader.utils",
        SimpleNamespace(search_process=lambda workflow_id, workflow: None),
    )
    monkeypatch.setitem(
        sys.modules,
        "transpiler_mate.oci",
        SimpleNamespace(OrasAnnotationsTranspiler=lambda *args, **kwargs: None),
    )

    try:
        cli.oci_annotations.callback.__wrapped__(
            source=source,
            workflow_id="main",
            image_source=None,
            image_revision=None,
            output=tmp_path / "annotations.json",
        )
        assert False, "Expected ValueError"
    except ValueError as exc:
        assert "Process main does not exist" in str(exc)


def test_markdown_and_invenio_publish_commands(monkeypatch, tmp_path: Path) -> None:
    source = tmp_path / "workflow.cwl"
    attach = tmp_path / "asset.txt"
    source.write_text("cwlVersion: v1.2\n", encoding="utf-8")
    attach.write_text("asset", encoding="utf-8")

    rendered = []
    monkeypatch.setattr(
        cli,
        "markdown_transpile",
        lambda source, workflow_id, output_stream, code_repository: (
            rendered.append((source, workflow_id, code_repository)),
            output_stream.write("# doc"),
        ),
    )
    cli.markdown.callback(
        source=source, workflow_id="main", output=tmp_path, code_repository=None
    )
    assert rendered[0][1] == "main"
    assert (tmp_path / "main.md").exists()

    created = {}
    monkeypatch.setattr(cli, "MetadataManager", lambda _source: "manager")
    monkeypatch.setitem(
        sys.modules,
        "transpiler_mate.invenio",
        SimpleNamespace(
            InvenioMetadataTranspiler=lambda metadata_manager, invenio_base_url, auth_token: (
                SimpleNamespace(
                    create_or_update_process=lambda source, attach: (
                        created.update({"source": source, "attach": attach})
                        or "https://invenio.example.org/records/1"
                    )
                )
            )
        ),
    )

    cli.invenio_publish.callback.__wrapped__(
        source=source,
        base_url="https://invenio.example.org",
        auth_token="token",
        attach=(attach,),
    )

    assert created["source"] == source
    assert created["attach"] == (attach,)



# Copyright 2026 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from datetime import date

from transpiler_mate.metadata.software_application_models import (
    AuthorRole,
    CreativeWork,
    DefinedTerm,
    Organization,
    Person,
    SoftwareApplication,
)
from transpiler_mate.ogcapi_records import (
    OgcRecordsTranspiler,
    SCIENCE_KEYWORDS_TERM_SET,
)
from transpiler_mate.ogcapi_records.sciencekeywords import ScienceKeywordRecord


def _software_application() -> SoftwareApplication:
    publisher = Organization(name="Terradue")
    author = AuthorRole(
        roleName="Principal Investigator",
        author=Person(
            givenName="Ada",
            familyName="Lovelace",
            email="ada@example.org",
            identifier="https://orcid.org/0000-0002-1825-0097",
            affiliation=Organization(
                name="Terradue", identifier="https://ror.org/03yrm5c26"
            ),
        ),
    )

    return SoftwareApplication(
        name="Example Tool",
        description="Example description",
        dateCreated=date(2026, 3, 9),
        license=[
            CreativeWork(
                identifier="Apache-2.0", url="https://spdx.org/licenses/Apache-2.0.html"
            ),
            CreativeWork(identifier="MIT", url="https://spdx.org/licenses/MIT.html"),
        ],
        softwareVersion="1.2.3",
        softwareHelp=CreativeWork(name="Help", url="https://example.org/help"),
        publisher=publisher,
        author=author,
        keywords=[
            "earth-observation",
            DefinedTerm(
                inDefinedTermSet=str(SCIENCE_KEYWORDS_TERM_SET),
                termCode="kw-1",
            ),
            DefinedTerm(termCode="ignored"),
        ],
    )


def test_ogc_records_transpile_maps_keywords_themes_and_contacts(monkeypatch) -> None:
    monkeypatch.setattr(
        "transpiler_mate.ogcapi_records.KEYWORDS_INDEX",
        {
            "kw-1": ScienceKeywordRecord(
                category="EARTH SCIENCE",
                topic="ATMOSPHERE",
                term="AEROSOLS",
                identifier="kw-1",
            )
        },
    )
    monkeypatch.setattr("transpiler_mate.ogcapi_records.uuid.uuid4", lambda: "abc123")

    result = OgcRecordsTranspiler().transpile(_software_application())

    assert result["id"] == "urn:uuid:abc123"
    assert result["type"] == "Feature"
    assert result["properties"]["keywords"] == ["earth-observation"]
    assert result["properties"]["themes"][0]["scheme"] == str(SCIENCE_KEYWORDS_TERM_SET)
    assert result["properties"]["themes"][0]["concepts"][0]["id"] == "EARTH SCIENCE"
    assert result["properties"]["contacts"][0]["position"] == "Principal Investigator"
    assert result["properties"]["license"] == "Apache-2.0: MIT"
    assert result["links"][0]["href"] == "https://example.org/help"


def test_ogc_records_transpile_skips_unsupported_keywords(monkeypatch) -> None:
    app = _software_application()
    app.keywords = [DefinedTerm(termCode="x"), 42]
    monkeypatch.setattr("transpiler_mate.ogcapi_records.uuid.uuid4", lambda: "abc123")

    result = OgcRecordsTranspiler().transpile(app)

    assert "keywords" not in result["properties"]
    assert "themes" not in result["properties"]



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .datacite_4_6_models import (
    Affiliation,
    Creator,
    Contributor,
    ContributorType,
    DataCiteAttributes,
    Date,
    DateType,
    Description,
    DescriptionType,
    Identifier,
    NameIdentifier,
    NameType,
    Publisher,
    RelatedIdentifier,
    RelatedIdentifierType,
    RelationType,
    ResourceType,
    ResourceTypeGeneral,
    Right,
    Title,
)
from ..metadata import Transpiler
from ..metadata.software_application_models import (
    AuthorRole,
    CreativeWork,
    Organization,
    Person,
    ContributorRole,
    SoftwareApplication,
)
from datetime import date
from pydantic import AnyUrl
from typing import Any, Mapping
from urllib.parse import urlparse

import time
import uuid

__ROLES_MAPPING_: Mapping[AnyUrl, ContributorType] = {
    AnyUrl(
        "http://purl.org/spar/datacite/ContactPerson"
    ): ContributorType.CONTACT_PERSON,
    AnyUrl(
        "http://purl.org/spar/datacite/DataCollector"
    ): ContributorType.DATA_COLLECTOR,
    AnyUrl("http://purl.org/spar/datacite/DataCurator"): ContributorType.DATA_CURATOR,
    AnyUrl("http://purl.org/spar/datacite/DataManager"): ContributorType.DATA_MANAGER,
    AnyUrl("http://purl.org/spar/datacite/Distributor"): ContributorType.DISTRIBUTOR,
    AnyUrl("http://purl.org/spar/datacite/Editor"): ContributorType.EDITOR,
    AnyUrl(
        "http://purl.org/spar/datacite/HostingInstitution"
    ): ContributorType.HOSTING_INSTITUTION,
    AnyUrl("http://purl.org/spar/datacite/Other"): ContributorType.OTHER,
    AnyUrl("http://purl.org/spar/datacite/Producer"): ContributorType.PRODUCER,
    AnyUrl(
        "http://purl.org/spar/datacite/ProjectLeader"
    ): ContributorType.PROJECT_LEADER,
    AnyUrl(
        "http://purl.org/spar/datacite/ProjectManager"
    ): ContributorType.PROJECT_MANAGER,
    AnyUrl(
        "http://purl.org/spar/datacite/ProjectMember"
    ): ContributorType.PROJECT_MEMBER,
    AnyUrl(
        "http://purl.org/spar/datacite/RegistrationAgency"
    ): ContributorType.REGISTRATION_AGENCY,
    AnyUrl(
        "http://purl.org/spar/datacite/RegistrationAuthority"
    ): ContributorType.REGISTRATION_AUTHORITY,
    AnyUrl(
        "http://purl.org/spar/datacite/RelatedPerson"
    ): ContributorType.RELATED_PERSON,
    AnyUrl("http://purl.org/spar/datacite/Researcher"): ContributorType.RESEARCHER,
    AnyUrl(
        "http://purl.org/spar/datacite/ResearchGroup"
    ): ContributorType.RESEARCH_GROUP,
    AnyUrl("http://purl.org/spar/datacite/RightsHolder"): ContributorType.RIGHTS_HOLDER,
    AnyUrl("http://purl.org/spar/datacite/Sponsor"): ContributorType.SPONSOR,
    AnyUrl("http://purl.org/spar/datacite/Supervisor"): ContributorType.SUPERVISOR,
    AnyUrl("http://purl.org/spar/datacite/Translator"): ContributorType.TRANSLATOR,
    AnyUrl(
        "http://purl.org/spar/datacite/WorkPackageLeader"
    ): ContributorType.WORK_PACKAGE_LEADER,
}


def _to_contributor(author: Person | ContributorRole) -> Contributor:
    contributor_type = ContributorType.OTHER

    if isinstance(author, ContributorRole):
        if author.additional_type:
            contributor_type = __ROLES_MAPPING_.get(
                author.additional_type, ContributorType.OTHER
            )

        author = author.contributor

    contributor: Contributor = Contributor(
        contributor_type=contributor_type,
        name_type=NameType.PERSONAL,
        name=f"{author.family_name}, {author.given_name}",
        given_name=author.given_name,
        family_name=author.family_name,
    )

    _finalize(author=author, creator=contributor)

    return contributor


def _to_creator(author: Person | AuthorRole) -> Creator:
    if isinstance(author, AuthorRole):
        author = author.author

    creator: Creator = Creator(
        name_type=NameType.PERSONAL,
        name=f"{author.family_name}, {author.given_name}",
        given_name=author.given_name,
        family_name=author.family_name,
    )

    _finalize(author=author, creator=creator)

    return creator


def _finalize(author: Person | Organization, creator: Creator):
    if author.identifier:
        creator.name_identifiers = []
        for identifier in (
            author.identifier
            if isinstance(author.identifier, list)
            else [author.identifier]
        ):
            scheme, netloc, _, _, _, _ = urlparse(str(identifier))
            creator.name_identifiers.append(
                NameIdentifier(
                    name_identifier=str(identifier),
                    name_identifier_scheme=netloc.split(".")[0].upper(),
                    scheme_uri=f"{scheme}://{netloc}",
                )
            )

    if isinstance(author, Person):
        creator.affiliation = []
        for affiliation in (
            author.affiliation
            if isinstance(author.affiliation, list)
            else [author.affiliation]
        ):
            if affiliation.identifier:
                for identifier in (
                    affiliation.identifier
                    if isinstance(affiliation.identifier, list)
                    else [affiliation.identifier]
                ):
                    scheme, netloc, _, _, _, _ = urlparse(str(identifier))
                    creator.affiliation.append(
                        Affiliation(
                            affiliation_identifier=str(identifier),
                            affiliation_identifier_scheme=netloc.split(".")[0].upper(),
                            scheme_uri=f"{scheme}://{netloc}",
                        )
                    )


class DataCiteTranspiler(Transpiler):
    def transpile(self, metadata_source: SoftwareApplication) -> Mapping[str, Any]:
        return DataCiteAttributes(
            doi=metadata_source.identifier,
            types=ResourceType(
                resource_type=metadata_source.name,
                resourceTypeGeneral=ResourceTypeGeneral.SOFTWARE,
            ),
            identifiers=[
                Identifier(identifier_type="DOI", identifier=metadata_source.identifier)
                if metadata_source.identifier
                else Identifier(
                    identifier_type="URN", identifier=f"urn:uuid:{uuid.uuid4()}"
                )
            ],  # supply a fake required identifier if the DOI hasn't been associated yet
            related_identifiers=[
                RelatedIdentifier(
                    related_identifier=str(metadata_source.same_as),
                    related_identifier_type=RelatedIdentifierType.DOI,
                    relation_type=RelationType.IS_IDENTICAL_TO,
                    resource_type_general=ResourceTypeGeneral.SOFTWARE,
                )
            ]
            if metadata_source.same_as
            else [],
            titles=[Title(title=metadata_source.name)],
            descriptions=[
                Description(
                    description=metadata_source.description,
                    description_type=DescriptionType.TECHNICAL_INFO,
                )
            ],
            publisher=Publisher(name=metadata_source.publisher.name),
            publication_year=metadata_source.date_created.year,
            dates=[
                Date(
                    date=date.fromtimestamp(time.time()),
                    date_type=DateType.UPDATED,
                    date_information="New version release",
                )
            ],
            rights=[
                Right(
                    rights=metadata_source.license.name
                    if isinstance(metadata_source.license, CreativeWork)
                    else None,
                    rights_uri=metadata_source.license.url
                    if isinstance(metadata_source.license, CreativeWork)
                    else None,
                    rights_identifier=metadata_source.license.identifier
                    if isinstance(metadata_source.license, CreativeWork)
                    else None,
                    rights_identifier_scheme="SPDX",
                )
            ],
            creators=list(
                map(
                    _to_creator,
                    metadata_source.author
                    if isinstance(metadata_source.author, list)
                    else [metadata_source.author],
                )
            ),
            contributors=list(
                map(
                    _to_contributor,
                    metadata_source.contributor
                    if isinstance(metadata_source.contributor, list)
                    else [metadata_source.contributor],
                )
            )
            if metadata_source.contributor
            else None,
        ).model_dump(exclude_none=True, by_alias=True)
